// updated
// ginc_crafting
/*
	Crafting related functions
*/
// ChazM 12/15/05
// ChazM 1/31/06 Moved lots of functions out to their appropriate include files.
// ChazM 1/31/06 Updated/fixed prototypes
// ChazM 2/1/06 Added Alchemy and Distillation.  Renamed and reorganized numerous functions.
// ChazM 2/1/06 Added Wondrous Items
// ChazM 2/2/06 Added various 2DA support
// ChazM 2/2/06 Added ability to use either 2DA or variables depending on if VAR_REC_SET is set.  Reorganized functions.
// ChazM 3/23/06 Added ExecuteDistillation()
// ChazM 3/23/06 Changed Mold suffix to be a mold prefix to match the current data set
// ChazM 3/28/06 Updated crafting to require various feats and skills
// ChazM 3/29/06 Added CasterLevel to Magical/Wondrous Crafting, addded additional error codes 
// ChazM 3/30/06 Added SuccessNotify(), updated Error codes - string refs with real values
// ChazM 4/8/06 Added bIdentify param to CreateListOfItemsInInventory()
// ChazM 4/11/06 Added MakeRepeatedItemList()
// ChazM 4/18/06 Modified CreateMagicalRecipe(), GetTargetItem(), and DoMagicCrafting() to support using item category (instead of tag suffix list)
// ChazM 5/5/06 Changed GetRecipeElement() to GetRecipeIntElement() and GetRecipeStringElement()
// ChazM 5/30/06 Added SetEnchantedItemName() - used in DoMagicCrafting()
// ChazM 5/31/06 Modified SetEnchantedItemName()
// ChazM 5/31/06 added strrefs to SetEnchantedItemName()
// ChazM 7/14/06 Modified DoMagicCrafting() -  Renaming Item only applies to Magic weapons and armor
// ChazM 7/29/06 Fix error in DoMundaneCrafting()
// ChazM 8/1/06 Max item properties set to 3. Masterwork weapons can be renamed on each enchanting, everything else only on the first enchanting
// ChazM 8/11/06 Updated str refs in SetEnchantedItemName()
// ChazM 8/11/06 updated CreateDistillationRecipe() - correction
// ChazM 8/14/06 added GetExpandedTags(), updated GetSortedItemList() - fix for stacks of items
// ChazM 8/16/06 added various workbench identification functions
// ChazM 8/27/06 added temp crafting VFX
// PKM 08.28.06 put in final VFX
// ChazM 9/10/06 Modified ExecuteDistillation() - stacks can now be distilled, added fix for playing special effect
// ChazM 9/29/06 DoMagicCrafting() - everything can be renamed upon enchanting, unless it has a special var; refixed SetEnchantedItemName()

//USER CHANGES
// Dash 11/13/06 - Updated DoMagicCrafting to allow adding effects to gear other than arms and armor
// Dash 11/13/06 - Updated DoMundaneCrafting to allow crafting a single object to rename it
// Dash 12/02/06 - Cleaned and reorganized DoMagicalCrafting, DoMundaneCrafting, and GetIsPropertyUpgrade
// Dash 12/02/06 - Integrated various new features in the aformentioned crafting functions
// Dash			 - Added GetFirstEqipmentItem function to get enchantment targets from inventories
// Dash			 - Added GetPropertyIsUpgrade to evaluate if a similar property already ixists
// Dash			 - Added GetNumberPropertiesOfType to count up properties which should be free
// Dash			 - Added a section of functions to support the salvagaing system
// Dash			 - Added a section of functions to support the set item creation system


#include "x0_i0_stringlib"
#include "ginc_vars"
#include "ginc_item"
#include "ginc_debug"	
#include "ginc_param_const"

// TCC Constants
const int TCC_TYPE_WONDER		= -2;
const int TCC_TYPE_ANY			= -1;
const int TCC_TYPE_NONE			= 0;
const int TCC_TYPE_OTHER		= 1;
const int TCC_TYPE_MELEE		= 2;
const int TCC_TYPE_BOW			= 3;
const int TCC_TYPE_XBOW			= 4;
const int TCC_TYPE_SLING		= 5;
const int TCC_TYPE_AMMO			= 6;
const int TCC_TYPE_ARMOR		= 7;
const int TCC_TYPE_SHIELD		= 8;
const int TCC_TYPE_HEAD			= 9;
const int TCC_TYPE_BACK			= 10;
const int TCC_TYPE_WRISTS		= 11;
const int TCC_TYPE_NECK			= 12;
const int TCC_TYPE_WAIST		= 13;
const int TCC_TYPE_RING			= 14;
const int TCC_TYPE_FEET			= 15;
const int TCC_TYPE_INSTRUMENT	= 16;
const int TCC_TYPE_CONTAINER	= 17;

// ************************
// *** Constants
// ************************
const string VAR_REC_SET = "RecipiesSet"; // Global Var

const string MAGICAL_RECIPE_PREFIX 		= "MAG";
const string MUNDANE_RECIPE_PREFIX		= "MUN";
const string ALCHEMY_RECIPE_PREFIX		= "ALC";
const string ALCHEMY_RECIPE_SUFFIX		= "ALC"; 
const string DISTILLATION_RECIPE_PREFIX	= "DIS";
const string DISTILLATION_RECIPE_SUFFIX = "DIS";

														// Used by:
const string VAR_RECIPE_COUNT   		= "_COUNT_";	// magical/wondrous, mundane, alchemy, distillation, 
//const string VAR_RECIPE_REAGENTS   		= "_LIST_";		// magical/wondrous, mundane, alchemy, distillation, 
//const string VAR_RECIPE_TAGS       		= "_TAGS_";		// magical only
//const string VAR_RECIPE_EFFECTS    		= "_EFFECT_";	// magical only
//const string VAR_RECIPE_OUTPUT			= "_OUT_";		// wondrous, mundane, alchemy, distillation
//const string VAR_RECIPE_CRAFT_SKILL		= "_SKILL_";	// mundane only
//const string VAR_RECIPE_SKILL_LEVEL		= "_LEVEL_";	// mundane, alchemy, distillation

// apparently this has changed to a suffix.
//const string MOLD_SUFFIX				= "_mld";		// tag of weapon/armor molds must always have this suffix.
const string MOLD_PREFIX				= "n2_crft_mold";		// tag of weapon/armor molds must always have this suffix.

const string VAR_RECIPE_SPELLID_LIST	= "RECIPE_SPELLID_LIST";	// list of SpellID indexes
const string VAR_RECIPE_RESREF_LIST		= "RECIPE_RESREF_LIST";		// list of mold resref inexes

const string VAR_ROW_NUMBER 			= "ROW_NUMBER";				// 2DA Row Number
const string VAR_RECIPE_2DA_INDEXES		= "RECIPE_2DA_INDEXES";		// List of info for index 2DA

const string CRAFTING_2DA 				= "crafting";
const string COL_CRAFTING_CATEGORY 		= "CATEGORY";	// magical/wondrous, mundane, alchemy, distillation, 
const string COL_CRAFTING_REAGENTS 		= "REAGENTS";	// magical/wondrous, mundane, alchemy, distillation, 
const string COL_CRAFTING_TAGS     		= "TAGS";		// magical only
const string COL_CRAFTING_EFFECTS  		= "EFFECTS";	// magical only
const string COL_CRAFTING_OUTPUT		= "OUTPUT";		// wondrous, mundane, alchemy, distillation
const string COL_CRAFTING_CRAFT_SKILL	= "SKILL";		// mundane only
const string COL_CRAFTING_SKILL_LEVEL	= "LEVEL";		// magical/wondrous, mundane, alchemy, distillation

	
const string CRAFTING_INDEX_2DA 		= "crafting_index";
//const string COL_CRAFTING_CATEGORY 	= "CATEGORY";	// this col in both 2DA's
const string COL_CRAFTING_START_ROW		= "START_ROW";

// TODO: these need to be updated to the real string refs
// Error codes - string refs
const int ERROR_ITEM_NOT_DISTILLABLE                = 174285; //"No match found for this item"        
const int ERROR_MISSING_REQUIRED_MOLD               = 174286; //"No mold found"    
const int ERROR_RECIPE_NOT_FOUND                    = 174287; //"No match found for this spell/reagent combo" 
const int ERROR_TARGET_NOT_FOUND_FOR_RECIPE         = 174288; //"couldn't find object for effect to be put on."
const int ERROR_INSUFFICIENT_CASTER_LEVEL           = 174289; //"Not high enough level of spell caster to use this magical recipe"
const int ERROR_INSUFFICIENT_CRAFT_ALCHEMY_SKILL	= 174290;
const int ERROR_INSUFFICIENT_CRAFT_ARMOR_SKILL      = 174291;
const int ERROR_INSUFFICIENT_CRAFT_WEAPON_SKILL     = 174293;
const int ERROR_NO_CRAFT_WONDROUS_FEAT              = 174294;
const int ERROR_NO_CRAFT_MAGICAL_FEAT               = 174295;
const int OK_CRAFTING_SUCCESS                 		= 174296;
const int ERROR_TARGET_HAS_MAX_ENCHANTMENTS			= 182996; // “Item can not be further enchanted.”

const int ERROR_UNRECOGNIZED_MORTAR_USAGE 			= 183206 ; // The mortar & pestle must be used on an alchemist's workbench or on an item.
const int ERROR_UNRECOGNIZED_HAMMER_USAGE 			= 183205 ; // "The smith hammer must be used on a blacksmith's workbench."		

// this variable is stored on the owned character so that script "gui_name_enchanted_item" can retrieve a reference to the item.
const string VAR_ENCHANTED_ITEM_OBJECT 	= "EnchantedItemObject";

// Standard Workbench Tag Prefixes
const string TAG_WORKBENCH_PREFIX1 = "PLC_MC_WBENCH";
const string TAG_WORKBENCH_PREFIX2 = "PLC_MC_CBENCH";

// Alchemy Workbench tags
const string TAG_ALCHEMY_BENCH1 	= "alchemy_bench";
const string TAG_ALCHEMY_BENCH2 	= "PLC_MC_CBENCH01";
const string TAG_ALCHEMY_BENCH3 	= "alchemy";

const string VAR_ALCHEMY			= "WB_alchemy";

// Blacksmith Workbench tags
const string TAG_WORKBENCH1 		= "workbench";
const string TAG_WORKBENCH2 		= "PLC_MC_CBENCH02";
const string TAG_WORKBENCH3 		= "blacksmith";

const string VAR_BLACKSMITH 		= "WB_blacksmith";
				
// just like vector, but with ints
struct IntVector
{
    int x;
 	int y;
	int z;
};

// ************************
// *** Prototypes
// ************************
	
// data functions	
string GetMagicalRecipeVar(string sRecipeVar, int iSpellID);
string GetMundaneRecipeVar(string sRecipeVar, string sMoldResRef);
string GetAlchemyRecipeVar(string sRecipeVar);
string GetDistillationRecipeVar(string sRecipeVar);
string GetRecipeVar(string sRecipeType, string sRecipeVar, string sCategory);
string GetCraftingStringData(string sRecipeType, string sCategory, string sColumn, int iCount);
int GetCraftingIntData(string sRecipeType, string sCategory, string sColumn, int iCount);

// useful functions (may be moved elsewhere)
string MakeList(string sReagent1, string sReagent2="", string sReagent3="", string sReagent4="", string sReagent5="",
            	string sReagent6="", string sReagent7="", string sReagent8="", string sReagent9="", string sReagent10="");
string MakeRepeatedItemList(string sListElement, int iCount);	
string MakeEncodedEffect(int nPropID, int nParam1 = 0, int nParam2 = 0, int nParam3 = 0, int nParam4 = 0);
void ApplyEncodedEffectToItem(string sEncodedEffect, object oItem);
void CreateListOfItemsInInventory(string sItemTemplateList, object oTarget, int bIdentify=TRUE, int bMasterWork=FALSE);
void DestroyItemsInInventory(int iIncludeCategories = ITEM_CATEGORY_ALL, object oTarget=OBJECT_SELF);
int Search2DA(string s2DA, string sColumn, string sMatchElement, int iStartRow, int iEndRow);
string PadString(int iMinSize, string sValue);

// functions for creating recipes
void CreateMagicalRecipe(int iSpellId, string sEffect, int iItemCategory, string sReagentList, int iCasterLevel=1);
//void CreateMagicalRecipe(int iSpellId, string sEffect, string sTagSuffixList, string sReagentList, int iCasterLevel=1);
void CreateWondrousRecipe(int iSpellId, string sReagentList, string sOutputResRef, int iCasterLevel=1);
void CreateMundaneRecipe(string sMoldTag, int iCraftSkill, int iSkillLevel, string sReagentList, string sOutputResRef);
void CreateAlchemyRecipe(int iSkillLevel, string sReagentList, string sOutputResRef);
void CreateDistillationRecipe(int iSkillLevel, string sReagent, string sOutputResRefList);

// **** 2da output related functions
struct IntVector GetRowIndexes(string sCategory);
string GetRecipeStringElement(string sRecipeType, string sRecipeVar, string sCategory, int iCount);
int GetRecipeIntElement(string sRecipeType, string sRecipeVar, string sCategory, int iCount);
string FormatHeaderRow();
string FormatRecipeRow(string sRecipePrefix, string sCategory, int iCount);
int OutputRecipeSet(string sRecipePrefix, string sIndex);
int OutputRecipeType(string sRecipePrefix, string sIndexList);
int OutputRecipes();
string FormatIndexHeaderRow();
string FormatIndexRecipeRow(int iRow, string sCategory, string sRowIndex);
int OutputRecipeIndex();

// private (helper) functions for using recipes
string GetSortedItemList(int iIncludeCategories = ITEM_CATEGORY_ALL, object oTarget=OBJECT_SELF);
int GetInventoryRecipeMatch(string sRecipePrefix, string sIndex, int iIncludeCategories = ITEM_CATEGORY_ALL);
int GetRecipeMatch(string sSortedItemList, string sRecipePrefix, string sIndex);
struct IntVector GetRowIndexes(string sCategory);
int GetCrafting2DARecipeMatch(string sSortedItemList, string sRecipePrefix, string sCategory);
object GetFirstEquipmentItem(object oTarget=OBJECT_SELF);
object GetTargetItem(int iItemCategory, object oTarget=OBJECT_SELF);
string FindMundaneIndexTag(string sTagSuffix, object oObject=OBJECT_SELF);
int GetNumberPropertiesOfType(object oItem, int iPropertyType);
int GetPropertyIsUpgrade(object oItem, itemproperty ipProperty);
int GetPropSlotsUsed(object oItem);
int GetHasExcludedProp(object oItem,int iRecipeMatch);

// additional helper functions for salvaging
string GetSalvageEssence(int iIndex,int iGrade);
string GetSalvageGem(int iIndex,int iGrade);
int GetSalvageDC(int iIndex,int iGrade);
int GetSalvageIndex(itemproperty ipProperty);
int GetSalvageGrade(itemproperty ipProperty,int iIndex);

// additional helper functions for set items
void ConvertToSetItem(object oOrigenal, string sSetName, object oContainer, object oPC);
void AddSetProperty(object oItem, int iSlot, string sPropString, object oPC);
void SetSetPropertyFlag(object oItem, int iSlot);
int GetSetPropertyFlag(object oItem);
int GetNumberSetProps(object oItem);
void CopyItemProperties(object oSource, object oTarget);
string GetMaterialCode(object oItem);
int GetNumEquippedParts(string sSetName, object oPC);
void SetSetLevelProperties(object oPC, string sSetName, int iSetCount);
void ClearSetProperties(object oItem);
void ActivateSetProperty(object oItem,int iLevel);

// functions for revised enchantment targeting
object GetEnchantmentTarget(string sTagList, object oContainer);
object GetTargetOfTCCType(int iTCCType, object oContainer);
int GetTCCType(object oItem);

// functions for using recipes
void DoMagicCrafting(int iSpellID, object oPC);
void DoMundaneCrafting(object oPC);
void DoAlchemyCrafting(object oPC);
void DoDistillation(object oItem, object oPC);
void DoSalvage(object oItem, object oPC);
void ExecuteDistillation(int iSkillReq, object oItem, object oPC, string sItemTemplateList);
void SuccessNotify(object oPC, int iStrRef=OK_CRAFTING_SUCCESS);
void ErrorNotify(object oPC, int iErrorStrRef);
void SetEnchantedItemName(object oPC, object oItem);

// is workbench functions
int IsWorkbench(object oTarget);
int IsSmithWorkbench(object oTarget);
int IsAlchemyWorkbench(object oTarget);


// ************************
// *** Functions
// ************************

// workbenches should all have 1 of 2 standard prefixes
int IsWorkbench(object oTarget)
{
	int iObjType = GetObjectType(oTarget);
	if (iObjType != OBJECT_TYPE_PLACEABLE) // smith workbench must be a placeable
		return FALSE;

	string sTargetTag = GetTag(oTarget);
	string sTargetTagPrefix = GetStringLeft(sTargetTag, GetStringLength(TAG_WORKBENCH_PREFIX1));
	if ((sTargetTagPrefix == TAG_WORKBENCH_PREFIX1) || (sTargetTagPrefix == TAG_WORKBENCH_PREFIX2))
		return TRUE;
		
	return FALSE;		
}

// Needed by Smith Hammer which works on Smith Workbench
int IsSmithWorkbench(object oTarget)
{
	int iObjType = GetObjectType(oTarget);
	if (iObjType != OBJECT_TYPE_PLACEABLE) // smith workbench must be a placeable
		return FALSE;

	// smith workbench can be identified by it's tag or by a local variable		
	string sTargetTag = GetTag(oTarget);
	if ((sTargetTag == TAG_WORKBENCH1) || (sTargetTag == TAG_WORKBENCH2) || (sTargetTag == TAG_WORKBENCH3))
		return TRUE;
		
	if (GetLocalInt(oTarget, VAR_BLACKSMITH) == TRUE)
		return TRUE;
		
	return FALSE;		
}


// Needed by Mortar & Pestle which only works on Alchemy Workbench
int IsAlchemyWorkbench(object oTarget)
{
	int iObjType = GetObjectType(oTarget);
	if (iObjType != OBJECT_TYPE_PLACEABLE) // alchemy workbench must be a placeable
		return FALSE;

	// alchemy workbench can be identified by it's tag or by a local variable		
	string sTargetTag = GetTag(oTarget);
	if ((sTargetTag == TAG_ALCHEMY_BENCH1) || (sTargetTag == TAG_ALCHEMY_BENCH2) || (sTargetTag == TAG_ALCHEMY_BENCH3))
		return TRUE;
		
	if (GetLocalInt(oTarget, VAR_ALCHEMY) == TRUE)
		return TRUE;
		
	return FALSE;		
}


// Toglable function for printing debug messages and feedback
void output(string sText, object oTarget = OBJECT_SELF)
{
	PrettyMessage(sText);
    //PrintString(sText);
    //AssignCommand(oTarget, SpeakString(sText));
	
	if(GetGlobalInt("CraftSpam"))
	{
		SendMessageToPC(GetFirstPC(),sText);
	}
}

	

// ===========================================
// Data Functions
// ===========================================

// Recipe info is stored in globals with the following format:
// MAG_[LIST/TAGS/EFFECT]_<SpellID>_X - where X is the index
string GetMagicalRecipeVar(string sRecipeVar, int iSpellID)
{
	string sVarName = GetRecipeVar(MAGICAL_RECIPE_PREFIX, sRecipeVar, IntToString(iSpellID));
    return (sVarName);
}

// Recipe info is stored in globals with the following format:
// MUN_[LIST/...]_<MoldResRef>_X
string GetMundaneRecipeVar(string sRecipeVar, string sMoldResRef)
{
	string sVarName = GetRecipeVar(MUNDANE_RECIPE_PREFIX, sRecipeVar, sMoldResRef);
    return (sVarName);
}

// Recipe info is stored in globals with the following format:
// ALC_[LIST/...]_ALC_X
string GetAlchemyRecipeVar(string sRecipeVar)
{
	string sVarName = GetRecipeVar(ALCHEMY_RECIPE_PREFIX, sRecipeVar, ALCHEMY_RECIPE_SUFFIX);
    return (sVarName);
}

string GetDistillationRecipeVar(string sRecipeVar)
{
	string sVarName = GetRecipeVar(DISTILLATION_RECIPE_PREFIX, sRecipeVar, DISTILLATION_RECIPE_SUFFIX);
    return (sVarName);
}

	
// Get Recipe Variable Name (used w/ SetGlobalArray*())
string GetRecipeVar(string sRecipeType, string sRecipeVar, string sCategory)
{
    string sVarName = sRecipeType + sRecipeVar + sCategory + "_";
	// tbd: should probably do a check to ensure string length not to long.
    return (sVarName);
}

// Get Specific RecipeElement from global var
string GetRecipeStringElement(string sRecipeType, string sRecipeVar, string sCategory, int iCount)
{
    string sVarRecipeList = GetRecipeVar(sRecipeType, sRecipeVar, sCategory);
    return (GetGlobalArrayString(sVarRecipeList, iCount));
}

// Get Specific RecipeElement from global var
int GetRecipeIntElement(string sRecipeType, string sRecipeVar, string sCategory, int iCount)
{
    string sVarRecipeList = GetRecipeVar(sRecipeType, sRecipeVar, sCategory);
    return (GetGlobalArrayInt(sVarRecipeList, iCount));
}

int UsingVariables()
{	
	return (GetGlobalInt(VAR_REC_SET));
}

// unified crafting data retrieval from global variables or 2da files.
// If globals variables are set then we use global vars.
// sRecipeType 	- var	- magical, mundane, etc - used as prefix of var name
// sCategory	- var	- spell id or other info used to determine var name 
// sColumn		- var/2da 	- the column of data we want (reagents, effects, etc.)
// iCount		- var/2da	- the count for var name -or- the index into the crafting.2da
string GetCraftingStringData(string sRecipeType, string sCategory, string sColumn, int iCount)
{
	string sStringData;
	if (UsingVariables())
	{
    	//string sGlobalArrayVar = GetRecipeVar(sRecipeType, sColumn, sCategory);
    	//sStringData = GetGlobalArrayString(sGlobalArrayVar, iCount);
		sStringData = GetRecipeStringElement(sRecipeType, sColumn, sCategory, iCount);
		
	}
	else // 2da version:
		sStringData = Get2DAString(CRAFTING_2DA, sColumn, iCount);

	return(sStringData);
}

int GetCraftingIntData(string sRecipeType, string sCategory, string sColumn, int iCount)
{
	int iIntData;		
	if (UsingVariables())
		iIntData = GetRecipeIntElement(sRecipeType, sColumn, sCategory, iCount);
	else // 2da version:
		iIntData = StringToInt(Get2DAString(CRAFTING_2DA, sColumn, iCount));

	return(iIntData);
	
	// return (StringToInt(GetCraftingStringData(sRecipeType, sCategory, sColumn, iCount)));
}

// ===========================================
// Useful funcs
// ===========================================

// Set up a list with up to 10 elements
// this list is simply a comma delimited string
// First element is required.
string MakeList(string sReagent1, string sReagent2="", string sReagent3="", string sReagent4="", string sReagent5="",
             	string sReagent6="", string sReagent7="", string sReagent8="", string sReagent9="", string sReagent10="")
{
    string sReagentList;

    sReagentList = sReagent1;
    sReagentList += FormListElement(sReagent2);
    sReagentList += FormListElement(sReagent3);
    sReagentList += FormListElement(sReagent4);
    sReagentList += FormListElement(sReagent5);
    sReagentList += FormListElement(sReagent6);
    sReagentList += FormListElement(sReagent7);
    sReagentList += FormListElement(sReagent8);
    sReagentList += FormListElement(sReagent9);
    sReagentList += FormListElement(sReagent10);

    return (sReagentList);
}

// will create a repeated list of items.
string MakeRepeatedItemList(string sListElement, int iCount)	
{
    string sList;
	int i;
	
	if (iCount >= 1)		
		sList = sListElement;

	for (i=2; i<=iCount; i++)		
    	sList += FormListElement(sListElement);
	

   	return (sList);
}


	
// Property ID required.  See function IPSafeAddItemProperty() in x2_inc_itemprop for supported props and params
// Creates a list containing the property and params for the effect to apply.
string MakeEncodedEffect(int nPropID, int nParam1 = 0, int nParam2 = 0, int nParam3 = 0, int nParam4 = 0)
{
    string sRecipeEffect;

    sRecipeEffect = IntToString(nPropID);
    sRecipeEffect += FormListElement(IntToString(nParam1));
    sRecipeEffect += FormListElement(IntToString(nParam2));
    sRecipeEffect += FormListElement(IntToString(nParam3));
    sRecipeEffect += FormListElement(IntToString(nParam4));

    return (sRecipeEffect);
}



// Apply an encoded effect to an item (created w/ MakeEncodedEffect())
void ApplyEncodedEffectToItem(string sEncodedEffect, object oItem)
{

    int nPropID = GetIntParam(sEncodedEffect, 0);
    int nParam1 = GetIntParam(sEncodedEffect, 1);
    int nParam2 = GetIntParam(sEncodedEffect, 2);
    int nParam3 = GetIntParam(sEncodedEffect, 3);
    int nParam4 = GetIntParam(sEncodedEffect, 4);

    itemproperty ipProperty = IPGetItemPropertyByID(nPropID, nParam1, nParam2, nParam3, nParam4);
	
	// Never replace property type bonus spell slot
	int iPolicy = X2_IP_ADDPROP_POLICY_REPLACE_EXISTING;
	if (GetItemPropertyType(ipProperty) == 13)
		iPolicy = X2_IP_ADDPROP_POLICY_IGNORE_EXISTING;
	
	// Ignore subtype for VFX properties
	int bIgnoreSub = FALSE;
	if (GetItemPropertyType(ipProperty) == 83)
		bIgnoreSub = TRUE;

    IPSafeAddItemProperty(oItem, ipProperty, 0.0f, iPolicy, FALSE, bIgnoreSub);
}

// Create items in target container, based on a comma delimited list of items to be created
// bIdentify: -1 leave as default, FALSE (0) - set as not identified, TRUE (1) - set as identified.
// Added masterwork parameter and related code - Dash 12/02/06
void CreateListOfItemsInInventory(string sItemTemplateList, object oTarget, int bIdentify=TRUE, int bMasterWork=FALSE)
{
	int nPos = 0;
	string sItemTemplate = GetStringParam(sItemTemplateList, nPos);
	string sMasterTemplate;
	object oCreatedObject;
	while (sItemTemplate != "")
	{
		if (bMasterWork)
		{
			sMasterTemplate = sItemTemplate + "_mast";
			oCreatedObject = CreateItemOnObject(sMasterTemplate, oTarget);
			if (!GetIsObjectValid(oCreatedObject))
				oCreatedObject = CreateItemOnObject(sItemTemplate, oTarget);			
		}
		else
		{
			oCreatedObject = CreateItemOnObject(sItemTemplate, oTarget);
		}
		 
		if (bIdentify != -1) SetIdentified(oCreatedObject, bIdentify);
		int iStackMax = StringToInt(Get2DAString("baseitems","Stacking",GetBaseItemType(oCreatedObject)));
		SetItemStackSize(oCreatedObject,iStackMax);
		nPos++;
 		sItemTemplate = GetStringParam(sItemTemplateList, nPos);
	}
}


// Destroy all items that are not base type armor or weapon
void DestroyItemsInInventory(int iIncludeCategories = ITEM_CATEGORY_ALL, object oTarget=OBJECT_SELF)
{
    object oItem =  GetFirstItemInInventory(oTarget);

    while (GetIsObjectValid(oItem))
    {
        if (GetIsItemCategory(oItem, iIncludeCategories))
        {
            DestroyObject(oItem, 0.2f);
        }
        oItem = GetNextItemInInventory(oTarget);
    }
}

// returns row number of match or -1 if not found.
// searches from start row to endrow
// search stops if empty string is returned (may be due to file, column, or row is not found or entry is "****"
int Search2DA(string s2DA, string sColumn, string sMatchElement, int iStartRow, int iEndRow)
{
	int i;
	string sEntry;

	for (i = iStartRow; i <= iEndRow; i++)
	{
		sEntry = Get2DAString(s2DA, sColumn, i);	
		if (sEntry == sMatchElement)
			{
			return i;
			}
		if (sEntry == "")
			{
			return -1;
			}
	}

	return -1;
}


//                                          1         2         3         4         5         6         7         8         9         0   
//            				       1234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
const string BLANK_SPACE_100 	= "                                                                                                    ";

// pad string w/ space to MinSize
string PadString(int iMinSize, string sValue)
{
	if (sValue == "" )
		sValue = "****";

	sValue += " ";
	int iLength = GetStringLength(sValue);
	int iNumSpaces = iMinSize - iLength;
	sValue += GetStringLeft(BLANK_SPACE_100, iNumSpaces);
	return (sValue);
}


// ===========================================
// functions for creating recipes
// ===========================================

// Magical recipes always require a spellID which is used as an index to improve search speed
// MAG_COUNT_<SPELL>_ = array count
// MAG_LIST_<SPELL>_X = recipe reagents
// MAG_TAGS_<SPELL>_X = recipe Tag Suffix List
// MAG_EFFECT_<SPELL>_X = recipe Effect
// MAG_LEVEL_<SPELL>_X = caster level required
void CreateMagicalRecipe(int iSpellId, string sEffect, int iItemCategory, string sReagentList, int iCasterLevel=1)
{
	string sNewList = AppendGlobalList(VAR_RECIPE_SPELLID_LIST, IntToString(iSpellId), TRUE);
	//output("new spellid list:" + sNewList);
    string sVarRecipeCount = GetMagicalRecipeVar(VAR_RECIPE_COUNT, iSpellId);
	int iCount = ModifyGlobalInt(sVarRecipeCount, 1);

    string sVarRecipeReagents   = GetMagicalRecipeVar(COL_CRAFTING_REAGENTS, iSpellId);
    string sVarRecipeTags       = GetMagicalRecipeVar(COL_CRAFTING_TAGS, iSpellId);
    string sVarRecipeEffect     = GetMagicalRecipeVar(COL_CRAFTING_EFFECTS, iSpellId);
   	string sVarRecipeCasterLevel= GetMagicalRecipeVar(COL_CRAFTING_SKILL_LEVEL, iSpellId);

    sReagentList = Sort(sReagentList);
	//output("CreateMagicalRecipe: Add " + sReagentList);

    SetGlobalArrayString(sVarRecipeReagents, iCount, sReagentList);
    SetGlobalArrayInt(sVarRecipeTags, iCount, iItemCategory);
    SetGlobalArrayString(sVarRecipeEffect, iCount, sEffect);
    SetGlobalArrayInt(sVarRecipeCasterLevel, iCount, iCasterLevel);
	
//    output ("sVarRecipeEffect = " + sVarRecipeEffect);
//    output ("sEffect = " + sEffect);
}

// Wonderous recipes always require a spellID which is used as an index to improve search speed
// These recipes are stored together with magical weapon/armor recipes.
// MAG_COUNT_<SPELL>_ = array count
// MAG_LIST_<SPELL>_X = recipe reagents
// MAG_OUT_<SPELL>_X = output
// MAG_LEVEL_<SPELL>_X = caster level required
void CreateWondrousRecipe(int iSpellId, string sReagentList, string sOutputResRef, int iCasterLevel=1)
{
	string sNewList = AppendGlobalList(VAR_RECIPE_SPELLID_LIST, IntToString(iSpellId), TRUE);
	//output("new spellid list:" + sNewList);
    string sVarRecipeCount = GetMagicalRecipeVar(VAR_RECIPE_COUNT, iSpellId);
	int iCount = ModifyGlobalInt(sVarRecipeCount, 1);

    string sVarRecipeReagents   = GetMagicalRecipeVar(COL_CRAFTING_REAGENTS, iSpellId);
    string sVarRecipeOutput   	= GetMagicalRecipeVar(COL_CRAFTING_OUTPUT, iSpellId);
   	string sVarRecipeCasterLevel= GetMagicalRecipeVar(COL_CRAFTING_SKILL_LEVEL, iSpellId);

    sReagentList = Sort(sReagentList);
	//output("CreateWondrousRecipe: Add " + sReagentList);

    SetGlobalArrayString(sVarRecipeReagents, iCount, sReagentList);
    SetGlobalArrayString(sVarRecipeOutput, iCount, sOutputResRef);
    SetGlobalArrayInt(sVarRecipeCasterLevel, iCount, iCasterLevel);
}


// Mundane recipes always require a mold which is used as an index to improve search speed
// MUN_COUNT_<MOLD>_ = array count
// MUN_LIST_<MOLD>_X = recipe reagents
// MUN_OUT_<MOLD>_X = output
// MUN_SKILL_<MOLD>_X = skill to be used (int)
// MUN_LEVEL_<MOLD>_X = Level of skill required
void CreateMundaneRecipe(string sMoldTag, int iCraftSkill, int iSkillLevel, string sReagentList, string sOutputResRef)
{
	string sNewList = AppendGlobalList(VAR_RECIPE_RESREF_LIST, sMoldTag, TRUE);
	//output("new resref list:" + sNewList);
	//output("new resref list:" + AppendGlobalList(VAR_RECIPE_RESREF_LIST, sMoldTag, TRUE));
    string sVarRecipeCount = GetMundaneRecipeVar(VAR_RECIPE_COUNT, sMoldTag);
	int iCount = ModifyGlobalInt(sVarRecipeCount, 1);

    string sVarRecipeReagents   = GetMundaneRecipeVar(COL_CRAFTING_REAGENTS, sMoldTag);
    string sVarRecipeOutput   	= GetMundaneRecipeVar(COL_CRAFTING_OUTPUT, sMoldTag);
    string sVarRecipeCraftSkill = GetMundaneRecipeVar(COL_CRAFTING_CRAFT_SKILL, sMoldTag);
    string sVarRecipeSkillLevel	= GetMundaneRecipeVar(COL_CRAFTING_SKILL_LEVEL, sMoldTag);

	sReagentList += FormListElement(sMoldTag); // add mold Res Ref to the list of reagents
    sReagentList = Sort(sReagentList);
	//output("CreateMundaneRecipe: Add recipe [" + IntToString(iCount) + "] - " + sReagentList);

    SetGlobalArrayString(sVarRecipeReagents, iCount, sReagentList);
    SetGlobalArrayString(sVarRecipeOutput, iCount, sOutputResRef);
    SetGlobalArrayInt(sVarRecipeCraftSkill, iCount, iCraftSkill);
    SetGlobalArrayInt(sVarRecipeSkillLevel, iCount, iSkillLevel);
}


// Alchemy recipes have no index.  Note that this means searches will get slower as more and more 
// recipes are added, so it should be kept to less than 50 or so.
void CreateAlchemyRecipe(int iSkillLevel, string sReagentList, string sOutputResRef)
{
    string sVarRecipeCount = GetAlchemyRecipeVar(VAR_RECIPE_COUNT);
	int iCount = ModifyGlobalInt(sVarRecipeCount, 1);

    string sVarRecipeReagents   = GetAlchemyRecipeVar(COL_CRAFTING_REAGENTS);
    string sVarRecipeOutput   	= GetAlchemyRecipeVar(COL_CRAFTING_OUTPUT);
    string sVarRecipeSkillLevel	= GetAlchemyRecipeVar(COL_CRAFTING_SKILL_LEVEL);

    sReagentList = Sort(sReagentList);

    SetGlobalArrayString(sVarRecipeReagents, iCount, sReagentList);
    SetGlobalArrayString(sVarRecipeOutput, iCount, sOutputResRef);
    SetGlobalArrayInt(sVarRecipeSkillLevel, iCount, iSkillLevel);
}


// Distillation recipes have no index.
void CreateDistillationRecipe(int iSkillLevel, string sReagent, string sOutputResRefList)
{
    string sVarRecipeCount = GetDistillationRecipeVar(VAR_RECIPE_COUNT);
	int iCount = ModifyGlobalInt(sVarRecipeCount, 1);

    string sVarRecipeReagents   = GetDistillationRecipeVar(COL_CRAFTING_REAGENTS);
    string sVarRecipeOutput   	= GetDistillationRecipeVar(COL_CRAFTING_OUTPUT);
    string sVarRecipeSkillLevel	= GetDistillationRecipeVar(COL_CRAFTING_SKILL_LEVEL);

    SetGlobalArrayString(sVarRecipeReagents, iCount, sReagent);
    SetGlobalArrayString(sVarRecipeOutput, iCount, sOutputResRefList);
    SetGlobalArrayInt(sVarRecipeSkillLevel, iCount, iSkillLevel);
}

// ====================================================================================
// Output to 2DA
// ====================================================================================

// generate a row of info representing a recipe
string FormatHeaderRow()
{
	PrintString("2DA V2.0");
	PrintString(" ");
	SetGlobalInt(VAR_ROW_NUMBER, -1);

	string sOut = "";
	sOut += PadString(5, " ");   
	sOut += PadString(15, COL_CRAFTING_CATEGORY);   
	sOut += PadString(100,COL_CRAFTING_REAGENTS);   
	sOut += PadString(20, COL_CRAFTING_TAGS);
	sOut += PadString(20, COL_CRAFTING_EFFECTS);
	sOut += PadString(50, COL_CRAFTING_OUTPUT);
	sOut += PadString(10, COL_CRAFTING_CRAFT_SKILL);
	sOut += PadString(10, COL_CRAFTING_SKILL_LEVEL);
	PrintString(sOut);
	return(sOut);
}


// generate a row of info representing a recipe
string FormatRecipeRow(string sRecipePrefix, string sCategory, int iCount)
{
	//output ("*** *** FormatRecipeRow: sRecipePrefix: " + sRecipePrefix + " sCategory:" + sCategory + " iCount:" + IntToString(iCount));
	int iRow = ModifyGlobalInt(VAR_ROW_NUMBER, 1);
	string sOut = "";
	sOut += PadString(5, IntToString(iRow));   
	sOut += PadString(15, sCategory);   
	sOut += PadString(100, GetRecipeStringElement(sRecipePrefix, COL_CRAFTING_REAGENTS	, sCategory, iCount));   
	sOut += PadString(20, IntToString(GetRecipeIntElement(sRecipePrefix, COL_CRAFTING_TAGS    	, sCategory, iCount)));
	sOut += PadString(20, GetRecipeStringElement(sRecipePrefix, COL_CRAFTING_EFFECTS    , sCategory, iCount));
	sOut += PadString(50, GetRecipeStringElement(sRecipePrefix, COL_CRAFTING_OUTPUT		, sCategory, iCount));
	sOut += PadString(10, IntToString(GetRecipeIntElement(sRecipePrefix, COL_CRAFTING_CRAFT_SKILL, sCategory, iCount)));
	sOut += PadString(10, IntToString(GetRecipeIntElement(sRecipePrefix, COL_CRAFTING_SKILL_LEVEL, sCategory, iCount)));
	PrintString(sOut);
	return(sOut);
}


// output all recipes for sRecipePrefix of a specific sIndex
int OutputRecipeSet(string sRecipePrefix, string sIndex)
{
	//AppendGlobalList(VAR_RECIPE_2DA_INDEXES, IntToString(GetGlobalInt(VAR_ROW_NUMBER)));
	AppendGlobalList(VAR_RECIPE_2DA_INDEXES, sIndex);
	AppendGlobalList(VAR_RECIPE_2DA_INDEXES, IntToString(1+GetGlobalInt(VAR_ROW_NUMBER)));

	//output ("*** OutputRecipeSet: sRecipePrefix: " + sRecipePrefix + " sIndex:" + sIndex);
    int iCount = 1;
    string sVarRecipeList = GetRecipeVar(sRecipePrefix, COL_CRAFTING_REAGENTS, sIndex);
    string sRecipeList;
    int bMatch = FALSE;

    sRecipeList = GetGlobalArrayString(sVarRecipeList, iCount);
    //output ("sRecipeList[" + IntToString(iCount) + "] = " + sRecipeList);

    while (sRecipeList != "") 
    {
		FormatRecipeRow(sRecipePrefix, sIndex, iCount);
		iCount++;
		sRecipeList = GetGlobalArrayString(sVarRecipeList, iCount);
		//output ("sRecipeList[" + IntToString(iCount) + "] = " + sRecipeList);
    }
    return (iCount);
}

// output all recipes of specific type, using the list of indexes to find them all
int OutputRecipeType(string sRecipePrefix, string sIndexList)
{
	//output ("OutputRecipeType: sRecipePrefix: " + sRecipePrefix + " sIndexList:" + sIndexList);
	string sIndex;
    struct sStringTokenizer stTok = GetStringTokenizer(sIndexList, ",");
	// loop through recipe index list
    while (HasMoreTokens(stTok)) {
        stTok = AdvanceToNextToken(stTok);
        sIndex = GetNextToken(stTok);
		OutputRecipeSet(sRecipePrefix, sIndex);
    }
	return TRUE;
}


// Output all recipes
int OutputRecipes()
{
	PrintString("Save the following as 'crafting.2da'");
	PrintString("===================================================");

	FormatHeaderRow();
	OutputRecipeType(MAGICAL_RECIPE_PREFIX, GetGlobalString(VAR_RECIPE_SPELLID_LIST));  	
	OutputRecipeType(MUNDANE_RECIPE_PREFIX, GetGlobalString(VAR_RECIPE_RESREF_LIST));	
	OutputRecipeType(ALCHEMY_RECIPE_PREFIX, ALCHEMY_RECIPE_SUFFIX);
	OutputRecipeType(DISTILLATION_RECIPE_PREFIX, DISTILLATION_RECIPE_SUFFIX);

	PrintString("===================================================");
	return TRUE;
}

// generate a row of info representing a recipe
string FormatIndexHeaderRow()
{
	PrintString("2DA V2.0");
	PrintString(" ");

	string sOut = "";
	sOut += PadString(5, " ");   
	sOut += PadString(15, "CATEGORY");   
	sOut += PadString(10, "START_ROW");   
	PrintString(sOut);
	return(sOut);
}

// generate a row of info representing a recipe
string FormatIndexRecipeRow(int iRow, string sCategory, string sRowIndex)
{
	//output ("*** *** FormatRecipeRow: sRecipePrefix: " + sRecipePrefix + " sCategory:" + sCategory + " iCount:" + IntToString(iCount));
	string sOut = "";
	sOut += PadString(5, IntToString(iRow));   
	sOut += PadString(15, sCategory);   
	sOut += PadString(5, sRowIndex);   
	PrintString(sOut);
	return(sOut);
}


// Output all recipes
int OutputRecipeIndex()
{
	int iRow = 0;
	PrintString("Save the following as 'crafting_index.2da'");
	PrintString("===================================================");

	FormatIndexHeaderRow();
	string s2DAIndexList = GetGlobalString(VAR_RECIPE_2DA_INDEXES);
	string sCategory, sRowIndex;
    struct sStringTokenizer stTok = GetStringTokenizer(s2DAIndexList, ",");
	// loop through recipe index list
    while (HasMoreTokens(stTok)) {
        stTok = AdvanceToNextToken(stTok);
        sCategory = GetNextToken(stTok);

        stTok = AdvanceToNextToken(stTok);
        sRowIndex = GetNextToken(stTok);
		FormatIndexRecipeRow(iRow, sCategory, sRowIndex);
		iRow++;
    }
	
	PrintString("===================================================");
	return TRUE;
}


// ====================================================================================
// Helper Functions for Using Recipes
// ====================================================================================

// return a list of tags to represent a stack of items
// never starts or ends w/ a seperator, so can be use just like a single tag for list making
string GetExpandedTags(object oItem)
{
	string sTag = GetTag(oItem);
	int iStackSize = GetItemStackSize(oItem);
	
	if (iStackSize < 1)
		iStackSize = 1; // just in case...
		
	string sRet = MakeRepeatedItemList (sTag, iStackSize);
	return (sRet);
}
	
// get all items in specified categories and sort the tags into an alphabetical list
string GetSortedItemList(int iIncludeCategories = ITEM_CATEGORY_ALL, object oTarget=OBJECT_SELF)
{
    object oItem =  GetFirstItemInInventory(oTarget);
    string sItemList = "";
	
    while (GetIsObjectValid(oItem))
    {	
	
		output("Item: " + GetTag(oItem) + "Category: " + IntToString(GetItemCategory(oItem)) + " (Match Category: " + IntToString(iIncludeCategories) + ")");
	
        if (GetIsItemCategory(oItem, iIncludeCategories))
        {
            if (sItemList == "")
                sItemList += GetExpandedTags(oItem);
            else
                sItemList += FormListElement(GetExpandedTags(oItem));
        }
        oItem = GetNextItemInInventory(oTarget);
    }
    sItemList = Sort(sItemList);
	
    return (sItemList);
}

// Creates a sorted list from an inventory, including only iIncludeCategories
// return index of the sorted list for this sRecipePrefix and sIndex, -1 if not found
int GetInventoryRecipeMatch(string sRecipePrefix, string sIndex, int iIncludeCategories = ITEM_CATEGORY_ALL)
{
    // list of reagent items in forge
    string sSortedItemList = GetSortedItemList(iIncludeCategories);
	int iRecipeMatch;

	output("Inventory: " + sSortedItemList + " Index: " + sIndex);
	
	iRecipeMatch = GetRecipeMatch(sSortedItemList, sRecipePrefix, sIndex);
	
	return (iRecipeMatch);
}

// return index of sSortedItemList for this sRecipePrefix and sIndex, -1 if not found
int GetRecipeMatch(string sSortedItemList, string sRecipePrefix, string sIndex)
{
    int iCount = 1;

	// do 2da look up instead if not using variables.
	if (!UsingVariables())
	{
		iCount = GetCrafting2DARecipeMatch(sSortedItemList, sRecipePrefix, sIndex);
		return (iCount);
	}

    string sVarRecipeList = GetRecipeVar(sRecipePrefix, COL_CRAFTING_REAGENTS, sIndex);
    string sRecipeList;
    int bMatch = FALSE;

    sRecipeList = GetGlobalArrayString(sVarRecipeList, iCount);

    while ((sRecipeList != "") && (bMatch == FALSE))
    {
        if (sSortedItemList == sRecipeList)
            bMatch = TRUE;
        else
        {
            iCount++;
            sRecipeList = GetGlobalArrayString(sVarRecipeList, iCount);
        }
    }
    if (bMatch == FALSE)
        iCount = -1;

    return (iCount);
}


// returns the start and end index rows for this category
struct IntVector GetRowIndexes(string sCategory)
{
	int iRow = -1;
	struct IntVector iv;
	iv.x = -1;
	iv.y = -1;
	
	int iIndexRow = Search2DA(CRAFTING_INDEX_2DA, COL_CRAFTING_CATEGORY, sCategory, 0, 150);
	if (iIndexRow == -1)
	{
		output ("Index Listing Not Found");
	}
	else
	{
		iv.x = StringToInt(Get2DAString(CRAFTING_INDEX_2DA, COL_CRAFTING_START_ROW, iIndexRow));
 		// ending row is 1 less than the value of the next start row
		iv.y = StringToInt(Get2DAString(CRAFTING_INDEX_2DA, COL_CRAFTING_START_ROW, iIndexRow+1))-1;
		if (iv.y == -1)
			iv.y = 10000; // go to the end.
	}
	return (iv);
}

// return index of sSortedItemList for this sRecipePrefix and sIndex, -1 if not found
int GetCrafting2DARecipeMatch(string sSortedItemList, string sRecipePrefix, string sCategory)
{
	int iCrafting2DARow = -1;
	struct IntVector iv = GetRowIndexes(sCategory);
		
	if (iv.x != -1)	
	{
		
		output("Index Listing Found: " + IntToString(iv.x) + " - " + IntToString(iv.y));
	
		iCrafting2DARow = Search2DA(CRAFTING_2DA, COL_CRAFTING_REAGENTS, sSortedItemList, iv.x, iv.y);	
	}
	
    return (iCrafting2DARow);
}

// Added by Dash 12/02/06
// Returns the first item of a type other than none or other
object GetFirstEquipmentItem(object oTarget=OBJECT_SELF)
{
	object oItem = OBJECT_INVALID;
	int iCategory = ITEM_CATEGORY_NONE;
	
	oItem =  GetFirstItemInInventory(oTarget);
	while (GetIsObjectValid(oItem))
	{
		iCategory = GetItemCategory(oItem);
		if (iCategory != ITEM_CATEGORY_NONE && iCategory != ITEM_CATEGORY_OTHER)
			break;
		oItem = GetNextItemInInventory(oTarget);
	}

	return oItem;
}

// todo - actually use taglist to verify object
// 
object GetTargetItem(int iItemCategory, object oTarget=OBJECT_SELF)
{
    object oItem = OBJECT_INVALID;
    //string sVarTagList = GetMagicalRecipeVar(COL_CRAFTING_TAGS, iSpellId);
    //string sTagList = GetGlobalArrayString(sVarTagList, iCount);

    oItem =  GetFirstItemInInventory(oTarget);
    //string sItemList = "";

    while (GetIsObjectValid(oItem))
    {
        if (GetItemCategory(oItem) == iItemCategory)
            break;
        oItem = GetNextItemInInventory(oTarget);
    }

    return oItem;
}

// looks for first object in inventory with sTagSuffix and returns full tag.
string FindMundaneIndexTag(string sTagPrefix, object oObject=OBJECT_SELF)
{
    object oItem = GetFirstItemInInventory(oObject);
	int iItemCount=0;
	int bFound = FALSE;
	string sRet = "";
	int iPrefixSize = GetStringLength(sTagPrefix);
	
    while (GetIsObjectValid(oItem) && !bFound)
    {
       	if (GetStringLeft(GetTag(oItem), iPrefixSize) == sTagPrefix)
       	{
			bFound = TRUE;
			sRet = GetTag(oItem);		
       	}
		else
       		oItem =  GetNextItemInInventory(oObject);
    }
    return sRet;
}

// Count the number of item properties
int GetNumberItemProperties(object oItem)
{
	int iItemPropCount = 0;
	//Get the first itemproperty
	itemproperty ipLoop=GetFirstItemProperty(oItem);
	
	while (GetIsItemPropertyValid(ipLoop))
	{
		iItemPropCount++;
	   	ipLoop=GetNextItemProperty(oItem);
	}
	return (iItemPropCount);
}

// Count the number of item properties
// Used slot costs assigned to the property type
int GetPropSlotsUsed(object oItem)
{
	int iItemPropCount = 0;
	int iPropCost = 0;
	int iType;
	itemproperty ipLoop = GetFirstItemProperty(oItem);
	
	while (GetIsItemPropertyValid(ipLoop))
	{
		iType = GetItemPropertyType(ipLoop);
		iPropCost = StringToInt(Get2DAString("itempropdef","Slots",iType));
		iItemPropCount = iItemPropCount + iPropCost;
	   	ipLoop = GetNextItemProperty(oItem);
	}
	return (iItemPropCount);
}

// Searches the target item for a property similar to the one specified
// Similar properties are the same type, and subtype of applicable
// Similar properties may have different parameters beyond this
// Returns true if any similar property is found, else returns false
int GetPropertyIsUpgrade(object oItem, itemproperty ipProperty)
{
	// Immediatly return false if the new property is a bonus spell slot
	// Bonus spell slots are never concidered an upgrade
	if (GetItemPropertyType(ipProperty) == 13)
		return FALSE;
	
	itemproperty ipScan = GetFirstItemProperty(oItem);
	
	// Loop through properties on item
    while (GetIsItemPropertyValid(ipScan))
    {
		int iProp = GetItemPropertyType(ipScan);	

		// Check primary property types and duration types for match
		if ((iProp == GetItemPropertyType(ipProperty)) && (GetItemPropertyDurationType(ipScan) == GetItemPropertyDurationType(ipProperty)))
        {
			// Determin if sub-type should be checked
			string sSub = Get2DAString("itempropdef", "SubTypeResRef", iProp);
			
			// VFX properties must ignore subtypes as a special case
			if (GetItemPropertyType(ipProperty) == 83)
				return TRUE;
			
			if (!(sSub == ""))
			{
				// Check property sub-type for match			
				if (GetItemPropertySubType(ipScan) == GetItemPropertySubType(ipProperty))
				{
					// Properties' types, duration types, and sub-types are fully matched
					return TRUE;
				}
				// ELSE (Implied)
				// Properties' types match but not sub-types
				// Keep searching the rest of the item
			}
			// Properties' types match and sub type is irrelevent
			else return TRUE;
		}
		// Properties and/or durations don't match
		// Get the next property and keep looking
		ipScan = GetNextItemProperty(oItem);		
	}
	// Loop exited without finding a match
	return FALSE;
}

// Returns the number of properties of the given type on an object
int GetNumberPropertiesOfType(object oItem, int iPropertyType)
{
	itemproperty ipScan = GetFirstItemProperty(oItem);
	int iCount = 0;

	// Loop through properties on item
    while (GetIsItemPropertyValid(ipScan))
    {
		if (iPropertyType == GetItemPropertyType(ipScan))
			iCount = iCount + 1;
		ipScan = GetNextItemProperty(oItem);		
	}
	return iCount;
}


// ====================================================================================
// Additional helper functions for salvaging
// ====================================================================================

// Produce the name of the essence associated with the property
string GetSalvageEssence(int iIndex,int iGrade)
{
	if (iGrade == 1)
		return Get2DAString("tcc_salvage","ESSENCE1",iIndex);
	else if (iGrade == 2)
		return Get2DAString("tcc_salvage","ESSENCE2",iIndex);
	else if (iGrade == 3)
		return Get2DAString("tcc_salvage","ESSENCE3",iIndex);
	else if (iGrade == 4)
		return Get2DAString("tcc_salvage","ESSENCE4",iIndex);
	else
		return "";
}

// Produce the name of the gem associated with the property
string GetSalvageGem(int iIndex,int iGrade)
{
	if (iGrade == 1)
		return Get2DAString("tcc_salvage","GEM1",iIndex);
	else if (iGrade == 2)
		return Get2DAString("tcc_salvage","GEM2",iIndex);
	else if (iGrade == 3)
		return Get2DAString("tcc_salvage","GEM3",iIndex);
	else if (iGrade == 4)
		return Get2DAString("tcc_salvage","GEM4",iIndex);
	else
		return "";
}

// Produce the skill check DC associated with the property
int GetSalvageDC(int iIndex,int iGrade)
{
	// Get gloval Salvaging DC modifier from configuration file
	int iSalvageDCMod = StringToInt(Get2DAString("tcc_config","value1",0));

	if (iGrade == 1)
		return StringToInt(Get2DAString("tcc_salvage","SKILL1",iIndex)) + iSalvageDCMod;
	else if (iGrade == 2)
		return StringToInt(Get2DAString("tcc_salvage","SKILL2",iIndex)) + iSalvageDCMod;
	else if (iGrade == 3)
		return StringToInt(Get2DAString("tcc_salvage","SKILL3",iIndex)) + iSalvageDCMod;
	else if (iGrade == 4)
		return StringToInt(Get2DAString("tcc_salvage","SKILL4",iIndex)) + iSalvageDCMod;
	else
		return 0;
}

// Get the index of the salvage line associated with the property
// Return -1 if there is no such salvage line
int GetSalvageIndex(itemproperty ipProperty)
{
	int iCounter = 0;
	int iSalvageType;
	
	int iPropType = GetItemPropertyType(ipProperty);
	string sSalvageCat = Get2DAString("tcc_salvage", "PROPERTY", iCounter);

	// Loop through the salvage chart

	while (sSalvageCat != "")
	{
		iSalvageType = StringToInt(sSalvageCat);
		if (iSalvageType == iPropType)
			return iCounter;
		iCounter = iCounter + 1;
		sSalvageCat = Get2DAString("tcc_salvage", "PROPERTY", iCounter);
	}
	return -1;
}

// Gets the proper salvage grade for the given property
// Returns a value from 1 to 4, or 0 on error
int GetSalvageGrade(itemproperty ipProperty, int iIndex)
{
	int iRange = StringToInt(Get2DAString("tcc_salvage","RANGE",iIndex));
	
	// Immediatly return cases with no scale factor
	if (iRange == 1)
		return 1;
		
	// Otherwise, solve for the proper grade value
	int iScaler = GetItemPropertyCostTableValue(ipProperty);

	if (iScaler <= iRange / 4)
		return 1;
	else if (iScaler <= (2 * iRange) / 4)
		return 2;
	else if (iScaler <= (3 * iRange) / 4)
		return 3;
	else
		return 4;
}

// ====================================================================================
// Additional helper functions for creating set items
// ====================================================================================

// Creates a retaged anotated copy of the given object for the given set
void ConvertToSetItem(object oOrigenal, string sSetName, object oContainer, object oPC)
{
	// Reject plot items before going further
	if (GetPlotFlag(oOrigenal) || GetLocalInt(oOrigenal,"tcc_noset") == 1)
	{
		SendMessageToPC(oPC,GetName(oOrigenal) + "can't be made part of an item set.");
		return;
	}
	
	// If set retagging is on and this is already set tagged then reject
	if (StringToInt(Get2DAString("tcc_config","value1",1)) && GetLocalString(oOrigenal,"tcc_setname") != "")
	{
		SendMessageToPC(oPC,GetName(oOrigenal) + "is already part of an item set.");
		return;
	}
	
	// Copy the origenal item with the set item tag
	object oNew = CopyObject(oOrigenal,GetLocation(oOrigenal),oContainer,"tcc_setitem");
	
	// Copy all properties from origenal to new item
	CopyItemProperties(oOrigenal,oNew);

	// Tag the new item with masterwork variable if needed
	if (GetStringRight(GetTag(oOrigenal),5) == "_mast")
		SetLocalInt(oNew,"tcc_masterwork",1);
		
	// Tag the new item with material variable if needed
	string sMaterial = GetMaterialCode(oOrigenal);
	if (sMaterial != "")
		SetLocalString(oNew,"tcc_material",sMaterial);
		
	// Tag the new item with the set name string
	SetLocalString(oNew,"tcc_setname",sSetName);
	
	// Add set name to begining of item name string
	SetFirstName(oNew,sSetName + " " + GetName(oOrigenal));
	
	// Destroy the origenal item
	DestroyObject(oOrigenal);
	
}

// Adds variables denoting a property item associated with an item set
// The actual property will be created by an on-equip script based on these variables
// iSlot indicates the number of set parts required to activate the property, 2 - 5
// sPropString is a five param comma delimited prop definition as seen in crafting.2da
// Any existing entry in the specified slot will be overwritten without notice
void AddSetProperty(object oItem, int iSlot, string sPropString, object oPC)
{
	// Confirm that the item is part of a set
	if(GetLocalString(oItem,"tcc_setname") != "")
	{
		// Apply the proper variable setting
		string sVarName = "tcc_setprop_" + IntToString(iSlot);
		if(GetLocalString(oItem,sVarName) != "")
			DeleteLocalString(oItem,sVarName);
		SetLocalString(oItem,sVarName,sPropString);
		DeleteLocalInt(oItem,"tcc_setflag");
	}
	else
	{
		// Alert the player that item isn't in a set
		SendMessageToPC(oPC,"This item is not part of a set!");
		return;
	}	
}

// Sets variables on an item indicating next property added should be a set property
// The iSlot parameter indicates what slot the next applied effect will be linked to
void SetSetPropertyFlag(object oItem, int iSlot)
{
	SetLocalInt(oItem,"tcc_setflag",iSlot);
}

// Returns the set slot value for the next property to be added to an item
// Returns zero if the flag is not set, indicating a non-set normal property
int GetSetPropertyFlag(object oItem)
{
	return GetLocalInt(oItem,"tcc_setflag");
}

// Returns the number of set prop strings stored on oItem
int GetNumberSetProps(object oItem)
{
	int iCount = 2;
	int iTotal = 0;
	
	// Get maximum number of set props from global config
	int iMax = StringToInt(Get2DAString("tcc_config","value1",3));
	
	while (iCount <= iMax + 1)
	{
		if (GetLocalString(oItem,"tcc_setprop_" + IntToString(iCount)) != "")
			iTotal = iTotal + 1;
		iCount = iCount + 1;
	}
	return iTotal;
}

// Coverts all elligble items in inventory into set members
// Destroys all inelligable items as assumed reagents
// This function actualy just pops a dialog box and feads data to another script
// The rest of the work is done in the other script
void ConstructSet(object oPC, object oContainer)
{
	object oObj = GetOwnedCharacter(oPC);
	SetLocalObject(oObj, "tcc_constructset", oContainer);
	SetLocalObject(oObj, "tcc_enchanter", oPC);

	DisplayInputBox(oPC,0,"Give your set a unique name:","gui_tcc_constructset","",FALSE,"",181744,"",0,"","My Set");
}

// Coppies all properties from oSource to oTarget
// Observes the same upgrade policies used for adding enchantments to items
void CopyItemProperties(object oSource, object oTarget)
{
	itemproperty ipProp = GetFirstItemProperty(oSource);
	while (GetIsItemPropertyValid(ipProp))
	{
		IPSafeAddItemProperty(oTarget, ipProp, 0.0f, X2_IP_ADDPROP_POLICY_REPLACE_EXISTING, FALSE, FALSE);
		ipProp = GetNextItemProperty(oSource);
	}
}

// Returns the five character material type code from an item
string GetMaterialCode(object oItem)
{
	// If local variable is set, return that immediatly
	string sLocalMat = GetLocalString(oItem,"tcc_material");
	if (sLocalMat != "")
		return sLocalMat;
	
	// Parse the tag string for a five character code
	string sTag = GetTag(oItem);	
	int iLength = GetStringLength(sTag);
	string sSample;
	int iCount = 5;
		
	while (iCount <= iLength)
	{
		sSample = GetStringRight(GetStringLeft(sTag,iCount),5);
	
		if (sSample == "_drk_")
			return "_drk_";
		else if (sSample == "_mth_")
			return "_mth_";
		else if (sSample == "_slh_")
			return "_slh_";
		else if (sSample == "_uhh_")
			return "_uhh_";
		else if (sSample == "_wyh_")
			return "_wyh_";
		else if (sSample == "_dsk_")
			return "_dsk_";
		else if (sSample == "_shd_")
			return "_shd_";
		else if (sSample == "_slv_")
			return "_slv_";
		else if (sSample == "_cld_")
			return "_cld_";
		else if (sSample == "_ada_")
			return "_ada_";
		else if (sSample == "_zal_")
			return "_zal_";
		else if (sSample == "_rdh_")
			return "_rdh_";
		iCount = iCount + 1;
	}
	
	// Nothing was found, assume no particular material
	return "";
}

// Returns the number of equipped items of the specified set
int GetNumEquippedParts(string sSetName, object oPC)
{
	object oItem;
	int iSlot = 0;
	int iCount = 0;
	
	// Loop through equipped slots counting matching items
	while (iSlot <= 16)
	{
		oItem = GetItemInSlot(iSlot,oPC);
		if (oItem != OBJECT_INVALID)
		{
			if (GetLocalString(oItem,"tcc_setname") == sSetName)
				iCount = iCount + 1;
		}
		iSlot = iSlot + 1;
	}
	return iCount;
}

// Activates or deactivates set bonuses to the proper level
void SetSetLevelProperties(object oPC, string sSetName, int iSetCount)
{
	// Loop through equipped slots modifying matching items
	int iSlot = 0;	
	int iMaxSetProps = StringToInt(Get2DAString("tcc_config","value1",2));
	int iTick = 2;
	string sProp = "";
	object oItem;
	while (iSlot <= 16)
	{
		iTick = 2;
		oItem = GetItemInSlot(iSlot,oPC);
		if (oItem != OBJECT_INVALID)
		{
			if (GetLocalString(oItem,"tcc_setname") == sSetName)
			{
				ClearSetProperties(oItem);

				while (iTick <= iMaxSetProps + 1)
				{
					sProp = GetLocalString(oItem,"tcc_setprop_" + IntToString(iTick));
					if (iSetCount >= iTick && sProp != "")
						ActivateSetProperty(oItem,iTick);
					iTick = iTick + 1;
				}
			}
		}
		iSlot = iSlot + 1;
	}
}

// Removes all set properties from the specified object
void ClearSetProperties(object oItem)
{
	int iPropLevel = 2;
	int iMaxSetProps = StringToInt(Get2DAString("tcc_config","value1",2));
	
	string sPropString;
	int nPropID;
	int nParam1;
	int nParam2;
	int nParam3;
	int nParam4;
	itemproperty ipProp;
	itemproperty ipMatch;
	
	int iPropCostTable;
	int iPropCostTableValue;
	int iPropSubType;
	int iPropType;
	
	int iMatchCostTable;
	int iMatchCostTableValue;
	int iMatchSubType;
	int iMatchType;
	
	string sSub;
	
	// Loop through the possable set property variables on the item
	while (iPropLevel <= iMaxSetProps + 1)
	{
	
		// If a set property variable exists, remove the associated property
		sPropString = GetLocalString(oItem,"tcc_setprop_" + IntToString(iPropLevel));
		if (sPropString != "")
		{
			// Compose item property for matching from string
			nPropID = GetIntParam(sPropString, 0);
			nParam1 = GetIntParam(sPropString, 1);
			nParam2 = GetIntParam(sPropString, 2);
			nParam3 = GetIntParam(sPropString, 3);
			nParam4 = GetIntParam(sPropString, 4);
			ipProp = IPGetItemPropertyByID(nPropID, nParam1, nParam2, nParam3, nParam4);
			
			// Get identity values from string variable prop
			iPropCostTable = GetItemPropertyCostTable(ipProp);
			iPropCostTableValue = GetItemPropertyCostTableValue(ipProp);
			iPropSubType = GetItemPropertySubType(ipProp);
			iPropType = GetItemPropertyType(ipProp);
			
			// Grab the subtype resref string for the property in question
			sSub = Get2DAString("itempropdef", "SubTypeResRef", iPropType);
			
			// Get first item property from object
			ipMatch = GetFirstItemProperty(oItem);
			
			// Loop through properties looking for a matching one
			while (GetIsItemPropertyValid(ipMatch))
			{
				// Get identity values from item property in question
				iMatchCostTable = GetItemPropertyCostTable(ipMatch);
				iMatchCostTableValue = GetItemPropertyCostTableValue(ipMatch);
				iMatchSubType = GetItemPropertySubType(ipMatch);
				iMatchType = GetItemPropertyType(ipMatch);
				
				// Determin if sub-type should be overridden (ignored)
				// True if looking at VFX props or props with no sub-type
				// If so, simply set the variables to be equal
				if (iPropType == 83 || sSub == "")
					iPropSubType = iMatchSubType;	
			
				if (iPropCostTable == iMatchCostTable && 
					iPropCostTableValue == iMatchCostTableValue && 
					iPropSubType == iMatchSubType && 
					iPropType == iMatchType)
				{
					RemoveItemProperty(oItem,ipMatch);
				}
				ipMatch = GetNextItemProperty(oItem);
			}
		}
		iPropLevel = iPropLevel + 1;
	}
}

// Adds the set property (if any) of the specified level
void ActivateSetProperty(object oItem,int iLevel)
{
	string sPropString = GetLocalString(oItem,"tcc_setprop_" + IntToString(iLevel));
	
	SendMessageToPC(GetFirstPC(FALSE),"Installing (" + sPropString + ") on " + GetName(oItem) + " (Level: " + IntToString(iLevel) + ")");
	
	int nPropID = GetIntParam(sPropString, 0);
    int nParam1 = GetIntParam(sPropString, 1);
    int nParam2 = GetIntParam(sPropString, 2);
    int nParam3 = GetIntParam(sPropString, 3);
    int nParam4 = GetIntParam(sPropString, 4);
    itemproperty ipProp = IPGetItemPropertyByID(nPropID, nParam1, nParam2, nParam3, nParam4);

	IPSafeAddItemProperty(oItem, ipProp, 0.0f, X2_IP_ADDPROP_POLICY_KEEP_EXISTING, FALSE, FALSE);
}


// ====================================================================================
// Support function for the recipe exclusion system
// ====================================================================================

int GetHasExcludedProp(object oItem,int iRecipeMatch)
{
	string sExclusions = Get2DAString("crafting","EXCLUDE",iRecipeMatch);
	
	if (sExclusions == "")
		return FALSE;
	
	int iParamVal = GetIntParam(sExclusions,0);
	int iTicker = 0;
	
	while (iParamVal != 0)
	{
		if (GetNumberPropertiesOfType(oItem,iParamVal) >= 1)
			return TRUE;
		iTicker = iTicker + 1;
		iParamVal = GetIntParam(sExclusions,iTicker);
	}
	return FALSE;	
}

// ====================================================================================
// New enchantment tag handeleing system
// ====================================================================================

// Looks in oContainer for an item of any of the types in sTagList
object GetEnchantmentTarget(string sTagList, object oContainer)
{
	int iTicker = 0;
	int iTag = GetIntParam(sTagList,0);
	object oTarget;
	
	while (iTag != 0)
	{
		oTarget = GetTargetOfTCCType(iTag, oContainer);
		if (GetIsObjectValid(oTarget))
			return oTarget;
		iTicker = iTicker + 1;
		iTag = GetIntParam(sTagList,iTicker);
	}
	
	return OBJECT_INVALID;
}

// Looks in oContainer for an object of type iTCCType
object GetTargetOfTCCType(int iTCCType, object oContainer)
{
	object oItem =  GetFirstItemInInventory(oContainer);

	while (GetIsObjectValid(oItem))
    {
		int iObjectType = GetTCCType(oItem);
        if ((iObjectType == iTCCType) || 
			(iTCCType == -1 && iObjectType >= 2) || 
			(iTCCType == -2 && iObjectType >= 9 && iObjectType <= 15))
            return oItem;
        oItem = GetNextItemInInventory(oContainer);
    }
	
	return OBJECT_INVALID;
}

// Gets the TCC Type of oItem
int GetTCCType(object oItem)
{
	// If object is invalid or not an item
	if (!GetIsObjectValid(oItem))
		return TCC_TYPE_NONE;
		
	int iBaseType = GetBaseItemType(oItem);
	
	// If object is any type of melee weapon
	if (iBaseType == BASE_ITEM_ALLUSE_SWORD || 
		iBaseType == BASE_ITEM_BASTARDSWORD || 
		iBaseType == BASE_ITEM_BATTLEAXE || 
		iBaseType == BASE_ITEM_CBLUDGWEAPON || 
		iBaseType == BASE_ITEM_CGIANT_AXE ||
		iBaseType == BASE_ITEM_CGIANT_SWORD || 
		iBaseType == BASE_ITEM_CLUB || 
		iBaseType == BASE_ITEM_CPIERCWEAPON ||
		iBaseType == BASE_ITEM_CSLASHWEAPON || 
		iBaseType == BASE_ITEM_CSLSHPRCWEAP || 
		iBaseType == BASE_ITEM_DAGGER ||
		iBaseType == BASE_ITEM_DIREMACE ||
		iBaseType == BASE_ITEM_DOUBLEAXE ||
		iBaseType == BASE_ITEM_DWARVENWARAXE || 
		iBaseType == BASE_ITEM_FALCHION ||
		iBaseType == BASE_ITEM_FLAIL || 
		iBaseType == BASE_ITEM_GREATAXE || 
		iBaseType == BASE_ITEM_GREATSWORD || 
		iBaseType == BASE_ITEM_HALBERD || 
		iBaseType == BASE_ITEM_HANDAXE || 
		iBaseType == BASE_ITEM_HEAVYFLAIL || 
		iBaseType == BASE_ITEM_KAMA || 
		iBaseType == BASE_ITEM_KATANA || 
		iBaseType == BASE_ITEM_KUKRI || 
		iBaseType == BASE_ITEM_LIGHTFLAIL || 
		iBaseType == BASE_ITEM_LIGHTHAMMER || 
		iBaseType == BASE_ITEM_LIGHTMACE || 
		iBaseType == BASE_ITEM_LONGSWORD || 
		iBaseType == BASE_ITEM_MACE || 
		iBaseType == BASE_ITEM_MORNINGSTAR || 
		iBaseType == BASE_ITEM_QUARTERSTAFF || 
		iBaseType == BASE_ITEM_RAPIER || 
		iBaseType == BASE_ITEM_SCIMITAR || 
		iBaseType == BASE_ITEM_SCYTHE || 
		iBaseType == BASE_ITEM_SHORTSPEAR || 
		iBaseType == BASE_ITEM_SHORTSWORD || 
		iBaseType == BASE_ITEM_SICKLE || 
		iBaseType == BASE_ITEM_SPEAR || 
		iBaseType == BASE_ITEM_TWOBLADEDSWORD || 
		iBaseType == BASE_ITEM_WARHAMMER || 
		iBaseType == BASE_ITEM_WARMACE || 
		iBaseType == BASE_ITEM_WHIP || 
		iBaseType == 120)
		return TCC_TYPE_MELEE;
	
	// If object is a long or short bow
	if (iBaseType == BASE_ITEM_SHORTBOW || 
		iBaseType == BASE_ITEM_LONGBOW)
		return TCC_TYPE_BOW;
	
	// If object is a light or heavy crossbow
	if (iBaseType == BASE_ITEM_HEAVYCROSSBOW || 
		iBaseType == BASE_ITEM_LIGHTCROSSBOW)
		return TCC_TYPE_XBOW;
		
	// If object is a sling
	if (iBaseType == BASE_ITEM_SLING)
		return TCC_TYPE_SLING;
		
	// If object is any type of ammunition
	if (iBaseType == BASE_ITEM_ARROW || 
		iBaseType == BASE_ITEM_BOLT || 
		iBaseType == BASE_ITEM_BULLET || 
		iBaseType == BASE_ITEM_DART || 
		iBaseType == BASE_ITEM_SHURIKEN || 
		iBaseType == BASE_ITEM_THROWINGAXE)
		return TCC_TYPE_AMMO;
	
	// If object is any type of armor
	if (iBaseType == BASE_ITEM_ARMOR)
		return TCC_TYPE_ARMOR;
		
	// If object is any type of shield
	if (iBaseType == BASE_ITEM_SMALLSHIELD || 
		iBaseType == BASE_ITEM_LARGESHIELD || 
		iBaseType == BASE_ITEM_TOWERSHIELD)
		return TCC_TYPE_SHIELD;
		
	// If object is a hat, healm, crown, etc
	if (iBaseType == BASE_ITEM_HELMET)
		return TCC_TYPE_HEAD;
		
	// If object is a necklace, amulet, etc
	if (iBaseType == BASE_ITEM_AMULET)
		return TCC_TYPE_NECK;
		
	// If object is a cape, cloak, mantle, etc
	if (iBaseType == BASE_ITEM_CLOAK)
		return TCC_TYPE_BACK;
	
	// If object is bracers, bracelets, glvoes, etc
	if (iBaseType == BASE_ITEM_BRACER || 
		iBaseType == BASE_ITEM_GLOVES)
		return TCC_TYPE_WRISTS;
		
	// If object is a ring of any type
	if (iBaseType == BASE_ITEM_RING)
		return TCC_TYPE_RING;
		
	// If object is a belt, gerdle, sash, etc
	if (iBaseType == BASE_ITEM_BELT)
		return TCC_TYPE_WAIST;
		
	// If object is boots, sandles, shooes, etc
	if (iBaseType == BASE_ITEM_BOOTS)
		return TCC_TYPE_FEET;
		
	// If object is a musical instrument
	if (iBaseType == BASE_ITEM_DRUM || 
		iBaseType == BASE_ITEM_FLUTE || 
		iBaseType == BASE_ITEM_MANDOLIN)
		return TCC_TYPE_INSTRUMENT;
		
	// If object is a bag, sack, etc
	if (iBaseType == 131)
		return TCC_TYPE_CONTAINER;
		
	return TCC_TYPE_OTHER;
}

// ====================================================================================
// Functions for Using Recipes
// ====================================================================================

// SPELL CAST AT WORKSTATION
// This covers two types of crafting:
// 1. Item Enchanting requires a set of reagents and an item add the effect too.	
// 2. Wondrous Crafting, requires a set of reagents and creates a new item.	

void DoMagicCrafting(int iSpellID, object oPC)
{
	// Initialize the script's core variables
	string sIndexTag =  IntToString(iSpellID);
	object oItem = OBJECT_INVALID;	
	int iReqFeat;
	int iError;
	int iItemPropCount;
	
	int iFirstSetRecipe = StringToInt(Get2DAString("tcc_config","value1",32));
	int iMaxSetProps = StringToInt(Get2DAString("tcc_config","value1",2));
	int iLastSetRecipe = iFirstSetRecipe + iMaxSetProps;	
	
    // Check for a recipe matching the reagents
    int iRecipeMatch = GetInventoryRecipeMatch(MAGICAL_RECIPE_PREFIX, sIndexTag, ITEM_CATEGORY_OTHER);
	int bDisabled = StringToInt(Get2DAString("crafting","DISABLED",iRecipeMatch));
	if (iRecipeMatch == -1 || !bDisabled)
    {
		ErrorNotify(oPC, ERROR_RECIPE_NOT_FOUND);
        return;
    }

	// Collect the required recipe information
	string sItemTemplateList 	= GetCraftingStringData(MAGICAL_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_OUTPUT, iRecipeMatch);
	int iReqCasterLevel 		= GetCraftingIntData(MAGICAL_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_SKILL_LEVEL, iRecipeMatch);
	string sTagList 			= GetCraftingStringData(MAGICAL_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_TAGS, iRecipeMatch);
	
	// Determine if we're creating or enchanting
	int bItemCreation = TRUE;
	if (sItemTemplateList == "")
		bItemCreation = FALSE;
	
	// Check for a target item for enchanting
	if (!bItemCreation)
	{
		// Attempt to get a base item matching one of the tags
	    oItem = GetEnchantmentTarget(sTagList,OBJECT_SELF);
		
		// Validate that an item was succesfully retrieved
	    if (!GetIsObjectValid(oItem))
	    {
	        ErrorNotify(oPC, ERROR_TARGET_NOT_FOUND_FOR_RECIPE);
	        return;
	    }
	}
	
	// Check for required caster level on caster
    if(GetCasterLevel(oPC) < iReqCasterLevel && StringToInt(Get2DAString("tcc_config","value1",3)))
	{		
  		ErrorNotify(oPC, ERROR_INSUFFICIENT_CASTER_LEVEL);
		return;
	}
		
	// Check for the required feat on caster
	int iItemType = GetTCCType(oItem);
	if (iItemType == TCC_TYPE_MELEE || 
		iItemType == TCC_TYPE_ARMOR || 
		iItemType == TCC_TYPE_SHIELD ||
		iItemType == TCC_TYPE_BOW || 
		iItemType == TCC_TYPE_XBOW || 
		iItemType == TCC_TYPE_SLING ||
		iItemType == TCC_TYPE_AMMO)
	{
		if (!GetHasFeat(FEAT_CRAFT_MAGIC_ARMS_AND_ARMOR,oPC) && StringToInt(Get2DAString("tcc_config","value1",4)))
		{
			ErrorNotify(oPC,ERROR_NO_CRAFT_MAGICAL_FEAT);
			return;
		}
	}
	else
	{
		if (!GetHasFeat(FEAT_CRAFT_WONDROUS_ITEMS,oPC) && StringToInt(Get2DAString("tcc_config","value1",4)))
		{
			ErrorNotify(oPC,ERROR_NO_CRAFT_WONDROUS_FEAT);
			return;
		}
	}
	
	// Check for properties excluded by this one
	if (!bItemCreation && StringToInt(Get2DAString("tcc_config","value1",33)))
	{
		if (GetHasExcludedProp(oItem,iRecipeMatch))
		{
			SendMessageToPC(oPC,"This recipe can't be combined with one or more properties already on the item.");
			return;
		}
	}

	// Check for available slots on item for enchants
	if (!bItemCreation && StringToInt(Get2DAString("tcc_config","value1",5)))
	{
		int iPropLimit = StringToInt(Get2DAString("tcc_config","value1",7));
		int iPropDiscount = 0;
		int iPropBonus = 0;
			
		// Assign a bonus slot if the item is a masterwork
		string sTag = GetTag(oItem);
		if (GetStringRight(sTag,5) == "_mast" || GetLocalInt(oItem,"tcc_masterwork") == 1)
			iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config","value1",6));
		
		// Increse limit by material code if available
		string sType = "Wep";
		if (iItemType == TCC_TYPE_ARMOR || 
			iItemType == TCC_TYPE_SHIELD)
			sType = "Arm";
		if (iItemType == TCC_TYPE_BOW || 
			iItemType == TCC_TYPE_XBOW || 
			iItemType == TCC_TYPE_SLING)
			sType = "Ran";
		
		if (StringToInt(Get2DAString("tcc_config","value1",8)))
		{
			string sSample = GetMaterialCode(oItem);
				
			if (sSample == "_drk_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",9));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",9));
			}
			else if (sSample == "_mth_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",10));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",10));
			}
			else if (sSample == "_slh_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",11));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",11));
			}
			else if (sSample == "_uhh_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",12));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",12));
			}
			else if (sSample == "_wyh_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",13));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",13));
			}
			else if (sSample == "_dsk_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",14));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",14));
			}
			else if (sSample == "_shd_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",15));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",15));
			}
			else if (sSample == "_slv_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",16));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",16));
			}
			else if (sSample == "_cld_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",17));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",17));
			}
			else if (sSample == "_ada_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",18));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",18));
			}
			else if (sSample == "_zal_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",19));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",19));
			}
			else if (sSample == "_rdh_")
			{
				iPropBonus = iPropBonus + StringToInt(Get2DAString("tcc_config",sType + "_Bonus",20));
				iPropDiscount = iPropDiscount + StringToInt(Get2DAString("tcc_config",sType + "_Allow",20));
			}
		}	

		// Grant an extra enchantment slot if the item has limitations
		// Also grant slot credit to offset each limitation property
		int iLimitations = 0;
		int iLimitationSlots = 0;
		int iPropCost = 1;
		int iNumPropsOfType = 0;
		int iTick = 62;
		int bUseVariableCosts = StringToInt(Get2DAString("tcc_config","value1",34));
		
		while (iTick <= 65)
		{
			iPropCost = StringToInt(Get2DAString("itempropdef","Slots",iTick));
			iNumPropsOfType = GetNumberPropertiesOfType(oItem,iTick);

			iLimitations = iLimitations + iNumPropsOfType;
			iLimitationSlots = iLimitationSlots + (iNumPropsOfType * iPropCost);
			
			iTick = iTick + 1;
		}
		
		if (iLimitations > 0)
		{
			if(StringToInt(Get2DAString("tcc_config","value1",22)))
				if (!bUseVariableCosts)
					iPropDiscount = iPropDiscount + iLimitations;
				else
					iPropDiscount = iPropDiscount + iLimitationSlots;
			if(StringToInt(Get2DAString("tcc_config","value1",21)))
				iPropBonus = iPropBonus + 1;
		}
				
		// Grant enchantment slot credit for each existing light effect
		if(StringToInt(Get2DAString("tcc_config","value1",23)))
		{
			int iNumLights = GetNumberPropertiesOfType(oItem,44);
			iPropCost = StringToInt(Get2DAString("itempropdef","Slots",44));
			
			if (!bUseVariableCosts)
				iPropDiscount = iPropDiscount + iNumLights;
			else
				iPropDiscount = iPropDiscount + (iNumLights * iPropCost);
		}
		
		// Grant enchantment slot credit for each existing VFX property
		if(StringToInt(Get2DAString("tcc_config","value1",24)))
		{
			int iNumVFX = GetNumberPropertiesOfType(oItem,83);
			iPropCost = StringToInt(Get2DAString("itempropdef","Slots",83));
			
			if (!bUseVariableCosts)
				iPropDiscount = iPropDiscount + iNumVFX;
			else
				iPropDiscount = iPropDiscount + (iNumVFX * iPropCost);
		}
			
		// Look for an existing property being replaced
		int bUpgrade = TRUE;
		string sEffect = GetCraftingStringData(MAGICAL_RECIPE_PREFIX, IntToString(iSpellID), COL_CRAFTING_EFFECTS, iRecipeMatch);
		itemproperty ipProperty = IPGetItemPropertyByID(GetIntParam(sEffect,0),GetIntParam(sEffect,1),GetIntParam(sEffect,2),GetIntParam(sEffect,3),GetIntParam(sEffect,4));
		bUpgrade = GetPropertyIsUpgrade(oItem, ipProperty);
		
		// Check to see if property to be added is free
		int bFreeProp = FALSE;
		int iType = GetIntParam(sEffect,1);
		if ((iType == 62 || iType == 63 ||iType == 64 || iType == 65) && StringToInt(Get2DAString("tcc_config","value1",22)))
			bFreeProp = TRUE;
		if (iType == 44 && StringToInt(Get2DAString("tcc_config","value1",23)))
			bFreeProp = TRUE;
		if (iType == 83 && StringToInt(Get2DAString("tcc_config","value1",24)))
			bFreeProp = TRUE;
		if ((iRecipeMatch >= iFirstSetRecipe && iRecipeMatch <= iLastSetRecipe) && StringToInt(Get2DAString("tcc_config","value1",25)))
			bFreeProp = TRUE;
		if (bUseVariableCosts && StringToInt(Get2DAString("itempropdef","Slots",iType)) == 0)
			bFreeProp = TRUE;

		// Look for number of existing properties
		int iItemPropCount;
		if (StringToInt(Get2DAString("tcc_config","value1",34)))
			iItemPropCount = GetPropSlotsUsed(oItem);
		else
			iItemPropCount = GetNumberItemProperties(oItem);		
		
		// Add the number of potential props from set items
		if (!StringToInt(Get2DAString("tcc_config","value1",25)))
			iItemPropCount = iItemPropCount + GetNumberSetProps(oItem);

		// Perform final enchant slot check
		if (iItemPropCount - iPropDiscount >= iPropLimit + iPropBonus && !bUpgrade && !bFreeProp)
	    {
	        ErrorNotify(oPC, ERROR_TARGET_HAS_MAX_ENCHANTMENTS);
	        return;
	    }
	}

	// Charge experiance points if required
	object oOwnedPC = GetOwnedCharacter(oPC);
	int iXPCost = StringToInt(Get2DAString("crafting","XP",iRecipeMatch));
	if (StringToInt(Get2DAString("tcc_config","value1",35)) && iXPCost != 0)
	{
		int iBaseXP = GetXP(oOwnedPC);
		if (iBaseXP >= iXPCost)
		{
			SetXP(oOwnedPC, iBaseXP - iXPCost);
		}
		else
		{
			SendMessageToPC(oPC,"You don't have enough experiance to create that.");
			return;
		}
	}
	
	// Charge gold coins if required
	int iGPCost = StringToInt(Get2DAString("crafting","GP",iRecipeMatch));
	if (StringToInt(Get2DAString("tcc_config","value1",36)) && iGPCost != 0)
	{
		if (GetGold(oOwnedPC) >= iGPCost)
		{
			TakeGoldFromCreature(iGPCost,oOwnedPC,TRUE);
		}
		else
		{
			SendMessageToPC(oPC,"You don't have enough gold to create that.");
		}
	}

	// Destroy component items in inventory
    DestroyItemsInInventory(ITEM_CATEGORY_OTHER);
	
	// Create or enchant items as necesary
	if (bItemCreation)
	{	
		// Create one or more new wondrous objects
		CreateListOfItemsInInventory(sItemTemplateList, OBJECT_SELF);
	}
	else
	{	
		// If this is a set item related recipe, handle it
		if ((iRecipeMatch >= iFirstSetRecipe && iRecipeMatch <= iLastSetRecipe) || GetSetPropertyFlag(oItem) != 0)
		{
			// Set up a new group of set items
			if (iRecipeMatch == iFirstSetRecipe)
			{
				ConstructSet(oPC,OBJECT_SELF);
				SendMessageToPC(oPC,"Item set forged!");
				return;
			}
			
			// Prepare item to recieve properties
			if (iRecipeMatch > iFirstSetRecipe && iRecipeMatch <= iLastSetRecipe)
			{
				int iCounter = iFirstSetRecipe;
				int iTicker = 2;
				while (iCounter <= iLastSetRecipe)
				{
					if (iRecipeMatch == iCounter)
					{
						SetSetPropertyFlag(oItem,iTicker);
						SendMessageToPC(oPC,"Item prepared! The next property added will require" + IntToString(iTicker) + "set parts to activate!");
						return;
					}
					iCounter = iCounter + 1;
					iTicker = iTicker + 1;
				}
			}

			// If we're still here, add selected effect as a set item property
			int iSlot = GetSetPropertyFlag(oItem);
			string sPropString = GetCraftingStringData(MAGICAL_RECIPE_PREFIX, IntToString(iSpellID), COL_CRAFTING_EFFECTS, iRecipeMatch);
			AddSetProperty(oItem,iSlot,sPropString,oPC);
			SendMessageToPC(oPC,"Effect added as set property!");
			return;
		}
	
		// Otherwise apply a new enchantment effect to object
		else
		{
			string sEncodedEffect = GetCraftingStringData(MAGICAL_RECIPE_PREFIX, IntToString(iSpellID), COL_CRAFTING_EFFECTS, iRecipeMatch);
			ApplyEncodedEffectToItem(sEncodedEffect, oItem);
		}
	}

	// Play a visual effect on work bench and notify player
    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectVisualEffect(VFX_FNF_CRAFT_MAGIC), OBJECT_SELF);
	SuccessNotify(oPC);
	
	// Allow player to name enchanted item
	if (!bItemCreation)	
	{		
		SetEnchantedItemName(oPC, oItem);
	}		
}

// SMITH HAMMER USED AT WORK STATION
// This does one of two things:
// 1. Mundate item creation using a set of reagents and a specific skill
// 2. Allows user to rename any Weapon, Armor, Shield, or Wonderous Object

void DoMundaneCrafting(object oPC)
{
	// Item renaming feature moved to i_smithhammer_ac

	// ITEM CREATION SECTION
    // Check for the presence of a mold
	string sIndexTag = FindMundaneIndexTag(MOLD_PREFIX);
	if (sIndexTag == "")
    {
  		ErrorNotify(oPC, ERROR_MISSING_REQUIRED_MOLD);
        return;
    }

	// Check for the required materials
    int iRecipeMatch = GetInventoryRecipeMatch(MUNDANE_RECIPE_PREFIX, sIndexTag, ITEM_CATEGORY_OTHER);
    if (iRecipeMatch == -1)
    {
  		ErrorNotify(oPC, ERROR_RECIPE_NOT_FOUND);
        return;
    }
	
    // Check for the required skill level
	int iSkill = GetCraftingIntData(MUNDANE_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_CRAFT_SKILL, iRecipeMatch);
	int iSkillReq = GetCraftingIntData(MUNDANE_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_SKILL_LEVEL, iRecipeMatch);
	int iPCSkill = GetSkillRank(iSkill, oPC);		
	if (iPCSkill < iSkillReq)
	{	
		int iError = ERROR_INSUFFICIENT_CRAFT_WEAPON_SKILL;
		if (iSkill == SKILL_CRAFT_ARMOR)
			iError = ERROR_INSUFFICIENT_CRAFT_ARMOR_SKILL;
   		ErrorNotify(oPC, iError);
		return;
	}
	
	// Check skill for masterwork qualification
	int bMasterWork = FALSE;
	int iMWSkillReq = StringToInt(Get2DAString("tcc_config","value1",31));
	if (iPCSkill >= iSkillReq + iMWSkillReq && StringToInt(Get2DAString("tcc_config","value1",26)))
	{
		bMasterWork = TRUE;
		SendMessageToPC(oPC,"Your have created a masterpiece!");
	}	
	
	// Destroy reagent materials in workspace
    DestroyItemsInInventory(ITEM_CATEGORY_OTHER);

	// Create the product object in workspace
	string sItemTemplateList = GetCraftingStringData(MUNDANE_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_OUTPUT, iRecipeMatch);
	CreateListOfItemsInInventory(sItemTemplateList, OBJECT_SELF, TRUE, bMasterWork);

	// Play crafting effect and notify player
    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectVisualEffect(VFX_FNF_CRAFT_BLACKSMITH), OBJECT_SELF);	// Charles, the parameter passed to EffectVisualEffect needs to be changed at your leisure
	SuccessNotify(oPC);
}

// Mortar and Pestle used at Alchemy workstation
// Notes: 
// Alchemy crafting requires a set of reagents, a specific skill level in Alchemy, and a Mortar & Pestle to activate it. 
// Reagents can be of any item type.
void DoAlchemyCrafting(object oPC)
{
    // is there a match for this recipe?
	string sIndexTag = ALCHEMY_RECIPE_SUFFIX; // alchemy has no index

    int iRecipeMatch = GetInventoryRecipeMatch(ALCHEMY_RECIPE_PREFIX, sIndexTag, ITEM_CATEGORY_OTHER);
    output("iRecipeMatch = " + IntToString(iRecipeMatch));
    if (iRecipeMatch == -1)
    {
   		ErrorNotify(oPC, ERROR_RECIPE_NOT_FOUND);
        return;
    }

    // meet required crafting skill?
	int iSkillReq = GetCraftingIntData(DISTILLATION_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_SKILL_LEVEL, iRecipeMatch);
	int iPCSkill = GetSkillRank(SKILL_CRAFT_ALCHEMY, oPC);		
	if (iPCSkill < iSkillReq)
	{	// has insufficient skill
   		ErrorNotify(oPC, ERROR_INSUFFICIENT_CRAFT_ALCHEMY_SKILL);
		return;
	}
	
    DestroyItemsInInventory(ITEM_CATEGORY_OTHER);

	//CreateOutput(iRecipeMatch, OBJECT_SELF);
	string sItemTemplateList = GetCraftingStringData(ALCHEMY_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_OUTPUT, iRecipeMatch);
	CreateListOfItemsInInventory(sItemTemplateList, OBJECT_SELF);

    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectVisualEffect(VFX_FNF_CRAFT_ALCHEMY), OBJECT_SELF);	// Charles, the parameter passed to EffectVisualEffect needs to be changed at your leisure
	SuccessNotify(oPC);
}

// MORTAR AND PESTIL USED ON AN ITEM
// This converts the target item into enchanting components
void DoDistillation(object oItem, object oPC)
{
	// STANDARD DISTILLATION SECTION
	// Check for a specific distillation recipe
	string sItemTag = GetTag(oItem);
	string sIndexTag = DISTILLATION_RECIPE_SUFFIX;
	int iItemCategory = GetItemCategory(oItem);

	int iRecipeMatch = GetRecipeMatch(sItemTag, DISTILLATION_RECIPE_PREFIX, sIndexTag);
	if (iRecipeMatch != -1)
	{
		output("Distilling...");
		// Check for required level of Alchemy skill
		int iSkillReq = GetCraftingIntData(DISTILLATION_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_SKILL_LEVEL, iRecipeMatch);
		string sItemTemplateList = GetCraftingStringData(DISTILLATION_RECIPE_PREFIX, sIndexTag, COL_CRAFTING_OUTPUT, iRecipeMatch);
		ExecuteDistillation(iSkillReq, oItem, oPC, sItemTemplateList);
	}
	// ITEM SALVAGE SECTION
	// Check if target item is potentialy salvagable
	else if (iItemCategory != ITEM_CATEGORY_OTHER && iItemCategory != ITEM_CATEGORY_NONE && StringToInt(Get2DAString("tcc_config","value1",27)))
	{
		if (!GetPlotFlag(oItem))
		{
			output("Salvaging...");
			DoSalvage(oItem,oPC);
		}
		else
		{
			output("Plot items can't be salvaged...");
			SendMessageToPC(oPC,"You cannot salvage a plot item!");
		}
	}
	else
	{
		output("Nothing to see here...");
		ErrorNotify(oPC, ERROR_ITEM_NOT_DISTILLABLE);
	}
}

// SUPPORT FUNCTION FOR ITEM DISTILATION
// Looks over the target item and acumulates salvaged materials
// Ouputs salvaged materials in the same way as distilled items
void DoSalvage(object oItem, object oPC)
{
	int iSalvageDC = 0;
	int iMinSkill = 0;
	int iScanDC = 0;
	string sFailureProduct = "";
	string sSuccessProduct = "";
	string sEss = "";
	string sGem = "";
	int iSalvageIndex = -1;
	int iSalvageGrade;
	
	// Loop through properties
	itemproperty ipScan = GetFirstItemProperty(oItem);
	while (GetIsItemPropertyValid(ipScan))
    {
		// If valid for salvage, process
		iSalvageIndex = GetSalvageIndex(ipScan);
		if (iSalvageIndex != -1)
		{
			// Get the grade of the property
			iSalvageGrade = GetSalvageGrade(ipScan,iSalvageIndex);
			
			// Get the dificulty of the property
			// Dificulty is the highest DC returned
			iScanDC = GetSalvageDC(iSalvageIndex,iSalvageGrade);
			if (iScanDC > iSalvageDC)
				iSalvageDC = iScanDC;
				
			// Get the minimum skill of the property
			// Minimum skill is the lowesest DC returned
			if (iScanDC < iMinSkill || iMinSkill == 0)
				iMinSkill = iScanDC;			
				
			// Assemble property's salvage product
			if (!(sFailureProduct == ""))
				sFailureProduct = sFailureProduct + ",";
			if (!(sSuccessProduct == ""))
				sSuccessProduct = sSuccessProduct + ",";
			sEss = GetSalvageEssence(iSalvageIndex,iSalvageGrade);
			sGem = GetSalvageGem(iSalvageIndex,iSalvageGrade);
			sFailureProduct = sFailureProduct + sEss;
			sSuccessProduct = sSuccessProduct + sEss + "," + sGem;
		}
		// Get the next property and loop
		ipScan = GetNextItemProperty(oItem);
	}
		
	// Check for required minimum skill level
	iMinSkill = iMinSkill - 5;
	if (iMinSkill > GetSkillRank(SKILL_SPELLCRAFT,oPC) && StringToInt(Get2DAString("tcc_config","value1",28)))
	{
		SendMessageToPC(oPC,"Your Spellcraft skill is not high enough to salvage this item.");
		return;
	}
	
	// Check for the chance of distilling a stack
	int iStackSize = GetItemStackSize(oItem);
	int iStackMax = StringToInt(Get2DAString("baseitems","Stacking",GetBaseItemType(oItem)));

	int iRoll = d100(1);
	int iChance = FloatToInt(100 * (IntToFloat(iStackSize) / IntToFloat(iStackMax)));
	
	if(iStackMax > 1)
	{
		if (iRoll > iChance)
		{
			SendMessageToPC(oPC,"<color=turquoise>Salvage Stack:</color><color=blue> *Failure* (1d100 rolled " + IntToString(iRoll) + " Vs " + IntToString(iChance) + ")");
			SendMessageToPC(oPC,"There was not enough left of the stack to salvage anything.");
			DestroyObject(oItem);
			return;
		}	
		else
		{
			SendMessageToPC(oPC,"<color=turquoise>Salvage Stack:</color><color=blue> *Success* (1d100 rolled " + IntToString(iRoll) + " Vs " + IntToString(iChance) + ")");
		}
	}
				
	// Do the skill check and create the results
	int bSkillSuccess = FALSE;
	if (StringToInt(Get2DAString("tcc_config","value1",29)))
	{
		if (GetIsSkillSuccessful(oPC,SKILL_SPELLCRAFT,iSalvageDC))
			bSkillSuccess = TRUE;
	}
	if (StringToInt(Get2DAString("tcc_config","value1",29)) && bSkillSuccess)
	{
		DestroyObject(oItem);
		CreateListOfItemsInInventory(sSuccessProduct,oPC);
	}
	else
	{
		DestroyObject(oItem);
		CreateListOfItemsInInventory(sFailureProduct,oPC);
	}	
	
}

void ExecuteDistillation(int iSkillReq, object oItem, object oPC, string sItemTemplateList)
{
    // TODO: meet required crafting skill?
	int iPCSkill = GetSkillRank(SKILL_CRAFT_ALCHEMY, oPC);		
	if (iPCSkill < iSkillReq)
	{	// has insufficient skill
   		ErrorNotify(oPC, ERROR_INSUFFICIENT_CRAFT_ALCHEMY_SKILL);
		return;
	}

	int iStackSize = GetItemStackSize(oItem); // we can distill multiple objects at once.
	DestroyObject(oItem);
	int i;
	for (i=1; i<= iStackSize; i++)
	{
		CreateListOfItemsInInventory(sItemTemplateList, oPC);
	}
			
    ApplyEffectAtLocation(DURATION_TYPE_INSTANT, EffectVisualEffect(VFX_FNF_CRAFT_SELF), GetLocation(oPC));	// Charles, the parameter passed to EffectVisualEffect needs to be changed at your leisure
	SuccessNotify(oPC);
}	

void SuccessNotify(object oPC, int iStrRef=OK_CRAFTING_SUCCESS)
{
	SendMessageToPCByStrRef(oPC, iStrRef);
}

void ErrorNotify(object oPC, int iErrorStrRef)
{
	SendMessageToPCByStrRef(oPC, iErrorStrRef);
}	


//PEH-OEI 05/24/06
//This script function displays a text input box popup on the client of the
//player passed in as the first parameter.
//////
// oPC           - The player object of the player to show this message box to
// nMessageStrRef- The STRREF for the Message Box message. 
// sMessage      - The text to display in the message box. Overrides anything 
//               - indicated by the nMessageStrRef
// sOkCB         - The callback script to call if the user clicks OK, defaults
//               - to none. The script name MUST start with 'gui'
// sCancelCB     - The callback script to call if the user clicks Cancel, defaults
//               - to none. The script name MUST start with 'gui'
// bShowCancel   - If TRUE, Cancel Button will appear on the message box.
// sScreenName   - The GUI SCREEN NAME to use in place of the default message box.
//               - The default is SCREEN_STRINGINPUT_MESSAGEBOX 
// nOkStrRef     - The STRREF to display in the OK button, defaults to OK
// sOkString     - The string to show in the OK button. Overrides anything that
//               - nOkStrRef indicates if it is not an empty string
// nCancelStrRef - The STRREF to dispaly in the Cancel button, defaults to Cancel.
// sCancelString - The string to display in the Cancel button. Overrides anything
//				 - that nCancelStrRef indicates if it is anything besides empty string
// sDefaultString- The text that gets copied into the input area,
//				 - used as a default answer
/*
void DisplayInputBox( object oPC, int nMessageStrRef,
						string sMessage, string sOkCB="", 
                        string sCancelCB="", int bShowCancel=FALSE, 
                        string sScreenName="",
                        int nOkStrRef=0, string sOkString="",
                        int nCancelStrRef=0, string sCancelString="",
                        string sDefaultString="", string sVariableString="" );
*/

// oPC = creator of the enchanted item.
// oItem = the enchanted item			
void SetEnchantedItemName(object oPC, object oItem)
{
	int nMessageStrRef 		= 181743;
	string sMessage 		= ""; // "Please rename the item.";
	string sOkCB			= "gui_name_enchanted_item";
	string sCancelCB		= "";
	int bShowCancel			= FALSE; 
	string sScreenName		= "";
	int nOkStrRef			= 181744;
	string sOkString		= "";
	int nCancelStrRef		= 181745;
	string sCancelString	= "";
	string sDefaultString 	= GetFirstName(oItem);
	string sVariableString	= "";
	

	// the gui script will always run on the owned PC, regardless of who the player has possessed.
	object oObj = GetOwnedCharacter(oPC);
	SetLocalObject(oObj, VAR_ENCHANTED_ITEM_OBJECT, oItem);

	DisplayInputBox( oPC, nMessageStrRef, sMessage, sOkCB, sCancelCB, bShowCancel, 
                     sScreenName, nOkStrRef, sOkString, nCancelStrRef, sCancelString,
                     sDefaultString, sVariableString);
}

//void main() {}
		