// ginc_item
/*
    Item/Inventory/Store related functions
*/
// ChazM 9/29/05
// ChazM 10/12/05 - added Remember, Restore, and Give family of functions
// ChazM 1/20/06 - added EquipRandomArmorFromInventory(), ListInventory(), GetInventoryItemOfBaseType()
// ChazM 1/31/06 - included "x2_inc_itemprop", added GetItemCategory(), GetIsItemCategory(), GetIsWeapon(), GetIsShield(), GetIsArmorOrShield()  
// ChazM 3/15/06 - added TransferItem(), modified GiveEquippedItem(), GiveAllInventory(), GiveEverything()
// ChazM 6/27/06 - added UnequipSlot()
// ChazM 6/29/06 - fixed UnequipSlot()
// ChazM 8/8/06 - Moved GetState() & SetState() to ginc_math
// ChazM 8/18/06 - added GetSlotOfEquippedItem(), GetIsCreatureSlot()
// ChazM 8/30/06 - added N2_GetNPCEasyMark(), N2_AppraiseOpenStore()
// ChazM 9/29/06 - Modified GetItemCategory() - now uses IP functions

//USER MODS
// Dash 11/13/06 - Updated GetItemCategory and category list to allow return of type ITEM_CATEGORY_WONDER
// Dash 11/13/06 - Added the GetIsWonderous feature to support the new ITEM_CATEGORY_WONDER return above

#include "ginc_debug"	
#include "x2_inc_itemprop"
#include "ginc_math"

//void main(){}
	
const string LAST_SLOT_NUM	=	"LastSlotNum";
const string LAST_SLOT_OBJ	=	"LastSlotObj";

// item categories - weapon, armor/shield, other
const int ITEM_CATEGORY_NONE    		= 0;
const int ITEM_CATEGORY_WEAPON  		= 1;
const int ITEM_CATEGORY_ARMOR_SHIELD   	= 2;
const int ITEM_CATEGORY_OTHER   		= 4;
const int ITEM_CATEGORY_ALL   			= 7;
const int ITEM_CATEGORY_WONDER			= 8;
const int ITEM_CATEGORY_GLOVES			= 9;
const int ITEM_CATEGORY_RANGED			= 10;
const int ITEM_CATEGORY_XBOW			= 11;
const int ITEM_CATEGORY_SLING			= 12;
const int ITEM_CATEGORY_BOW				= 13;
const int ITEM_CATEGORY_AMMO			= 14;
const int ITEM_CATEGORY_ANY				= 15;

//-------------------------------------------------
// Function Prototypes
//-------------------------------------------------
	
// item functions
int GetIsItemEquipped(string sItemTag, object oCreature, int bExactMatch = TRUE);
int GetPartyMemberHasEquipped(string sItemTag, object oPartyMember, int bPCOnly=TRUE, int bExactMatch = TRUE);
int RemoveItem (object oCreature, string sItemTag);
void CreateItemOnFaction(object oPC, string sItemResRef, int bPCOnly=TRUE);
void EquipNewItem(object oTarget, string sItemResRef, int iEquipSlot);
void DoEquipmentUpgrade(string sTargetTag, string sItemResRef, int iEquipSlot);
void TransferItem(object oGiver, object oReceiver, object oInventoryItem);

void RememberEquippedItem(object oOwner, int iSlot);
void RememberEquippedItems(object oOwner);
void RestoreEquippedItem(object oOwner, object oInventoryItem, int bReset=TRUE);
void RestoreEquippedItems(object oOwner, int bReset=TRUE);

void UnequipSlot(object oOwner, int iSlot);
void GiveEquippedItem(object oGiver, object oReceiver, int iSlot);
void GiveAllEquippedItems(object oGiver, object oReceiver);
void GiveAllInventory(object oGiver, object oReceiver);
void GiveEverything(object oGiver, object oReceiver);

int GetItemCategory(object oItem);
int GetIsItemCategory(object oItem, int iItemCategory);
int GetIsWeapon(object oItem);
int GetIsShield(object oItem);
int GetIsArmorOrShield(object oItem);
int GetIsWonderous(object oItem);

int GetSlotOfEquippedItem(object oItem, object oCreature);
int GetIsCreatureSlot(int iSlot);


// *** Store related
int N2_GetNPCEasyMark(object oTarget);
void N2_AppraiseOpenStore(object oStore, object oPC, int nBonusMarkUp = 0, int nBonusMarkDown = 0);

//-------------------------------------------------
// Function Definitions
//-------------------------------------------------

// returns true if item equipped
//  sItemTag - tag of the item to look for
//  oPartyMember - creature whose equipment is to be searched.
//  bExactMatch - whether or not to search tags with exact or partial matches
int GetIsItemEquipped(string sItemTag, object oCreature, int bExactMatch = TRUE)
{
    object oItem;
    string sFoundTag;
	int nSlot;
	
    for (nSlot=0; nSlot<NUM_INVENTORY_SLOTS; nSlot++)
    {
        oItem = GetItemInSlot(nSlot, oCreature);

        if (TRUE == GetIsObjectValid(oItem))
		{
			sFoundTag = GetTag(oItem);

			if (FALSE == bExactMatch)
			{
				if (FindSubString(sFoundTag, sItemTag) != -1)
					return TRUE;
			}
			else
			{
				if (sFoundTag == sItemTag)
					return TRUE;
			}
        }
    }
    return FALSE;
}

// returns true if any party member has item equipped
//  sItemTag - tag of the item to look for
//  oPartyMember - Any member of the party
//  bPCOnly - Include only PC's or everyone in the party.  Default is only the PCs.
int GetPartyMemberHasEquipped(string sItemTag, object oPartyMember, int bPCOnly=TRUE, int bExactMatch = TRUE)
{
    object oMember = GetFirstFactionMember(oPartyMember, bPCOnly); // Needed GetFirstPC() to run through PC List

    while (oMember != OBJECT_INVALID)
    {
        if (GetIsItemEquipped(sItemTag, oMember, bExactMatch))
            return TRUE;
        oMember = GetNextFactionMember(oPartyMember, bPCOnly);
    }
    return FALSE;
}

// Take item from player inventory if possessed
// currently if stackable, takes whole stack.
// return true if object was available to be removed
int RemoveItem (object oCreature, string sItemTag)
{
	int iRet = FALSE;
    object oItemToTake = GetItemPossessedBy(oCreature, sItemTag);
    if(GetIsObjectValid(oItemToTake))
	{
        DestroyObject(oItemToTake);
		iRet = TRUE;
	}
	return iRet;
}


void CreateItemOnFaction(object oPC, string sItemResRef, int bPCOnly=TRUE)
{
    object oPartyMem = GetFirstFactionMember(oPC, bPCOnly);
    while (GetIsObjectValid(oPartyMem)) 
	{
		CreateItemOnObject(sItemResRef, oPC);
        oPartyMem = GetNextFactionMember(oPC, bPCOnly);
    }
}
	
		
// give oTarget specified equipment and put in specified slot
void EquipNewItem(object oTarget, string sItemResRef, int iEquipSlot)
{
	object oItem = CreateItemOnObject(sItemResRef, oTarget, 1);
	//Equip item (need time to allow item to appear)
	if (iEquipSlot >= 0)
		DelayCommand(0.5, AssignCommand(oTarget, ActionEquipItem(oItem, iEquipSlot)));
}

// give all w/ sTargetTag, specified equipment and put in specified slot
void DoEquipmentUpgrade(string sTargetTag, string sItemResRef, int iEquipSlot)
{
	int i = 0;
	object oTarget = GetObjectByTag(sTargetTag, i);
	while (GetIsObjectValid(oTarget))
	{
//		PrintString("Equiping NPC " + GetName(oTarget) + " - index " + IntToString(i));
		EquipNewItem(oTarget, sItemResRef, iEquipSlot);
		oTarget = GetObjectByTag(sTargetTag, ++i);
	}
	DebugMessage("Total number of creatures updated: " + IntToString(i));
}


// store info on item about current owner and slot equipped
void RememberEquippedItem(object oOwner, int iSlot)
{
	object oInventoryItem = GetItemInSlot(iSlot, oOwner);
	if (oInventoryItem != OBJECT_INVALID)
	{
		SetLocalInt(oInventoryItem, LAST_SLOT_NUM, iSlot);
		SetLocalObject(oInventoryItem, LAST_SLOT_OBJ, oOwner);
	}
}

// store info on all equipped items of owner
void RememberEquippedItems(object oOwner)
{
    int iSlot;
    for (iSlot = 0; iSlot < NUM_INVENTORY_SLOTS; iSlot++)
    {
		RememberEquippedItem(oOwner, iSlot);
    }
}


// restore item to slot last remembered
void RestoreEquippedItem(object oOwner, object oInventoryItem, int bReset=TRUE)
{
	int iSlot = GetLocalInt(oInventoryItem, LAST_SLOT_NUM);
	//object oOrigOwner = GetLocalObject(oInventoryItem, LAST_SLOT_OBJ);

	if (iSlot > 0)
	{
	    AssignCommand(oOwner, ActionEquipItem(oInventoryItem, iSlot));
		if (bReset)
		{
			DeleteLocalInt(oInventoryItem, LAST_SLOT_NUM);
			DeleteLocalObject(oInventoryItem, LAST_SLOT_OBJ);
		}
	}
}

// Restore all inventory items to equipment slot last remembered (if any)
void RestoreEquippedItems(object oOwner, int bReset=TRUE)
{
    object oInventoryItem = GetFirstItemInInventory(oOwner);
	object oLeftHandItem;
	int iSlot;
	
	while (oInventoryItem != OBJECT_INVALID)
    {
		iSlot = GetLocalInt(oInventoryItem, LAST_SLOT_NUM);
		
		//EPF 8/29/06 left hand must be equipped after right.  Otherwise the engine
		//    will shift it to the right hand slot automatically.
		if(iSlot != INVENTORY_SLOT_LEFTHAND)
		{
			RestoreEquippedItem(oOwner, oInventoryItem, bReset);
	        oInventoryItem = GetNextItemInInventory(oOwner);
		}
		else
		{
			oLeftHandItem = oInventoryItem;
		}
    }
	
	if(GetIsObjectValid(oLeftHandItem))
	{
		RestoreEquippedItem(oOwner, oLeftHandItem, bReset);
	}
}

// unequip item in specified slot, if any
void UnequipSlot(object oOwner, int iSlot)
{
	object oInventoryItem = GetItemInSlot(iSlot, oOwner);
	if (oInventoryItem != OBJECT_INVALID)
	{		
		AssignCommand(oOwner, ActionUnequipItem(oInventoryItem));
	}		
}


// passing around carried stuff.
void GiveEquippedItem(object oGiver, object oReceiver, int iSlot)
{	
	if (GetObjectType(oGiver) != OBJECT_TYPE_CREATURE)
	{
		PrettyError("GiveEquippedItem(): Equipped Item Giver must be a creature!");		
		return;
	}

	object oInventoryItem = GetItemInSlot(iSlot, oGiver);

	if (oInventoryItem != OBJECT_INVALID)
	{
		AssignCommand(oGiver, ActionUnequipItem(oInventoryItem));
		//AssignCommand(oReceiver, ActionTakeItem(oInventoryItem, oGiver));
        //AssignCommand(oGiver, ActionGiveItem(oInventoryItem, oReceiver));
		TransferItem(oGiver, oReceiver, oInventoryItem);
	}
}
	
void GiveAllEquippedItems(object oGiver, object oReceiver)
{	
    int iSlot;
    for (iSlot = 0; iSlot < NUM_INVENTORY_SLOTS; iSlot++)
    {
		GiveEquippedItem(oGiver, oReceiver, iSlot);
    }
}


// Transfer an item from giver to receiver.
// objects lack an action queue, so 
void TransferItem(object oGiver, object oReceiver, object oInventoryItem)
{
	// placeable give/take item commands are most dependable
	if (GetObjectType(oReceiver) == OBJECT_TYPE_PLACEABLE)
	{ // the receiver is a creature, so he can be assigned ActionTakeItem() commands to his queue
        AssignCommand(oReceiver, ActionTakeItem(oInventoryItem, oGiver));
	}
	else //if ((GetObjectType(oGiver) == OBJECT_TYPE_PLACEABLE)
	{ // if both are creatures, giving item better than taking since taking causes movement.
        AssignCommand(oGiver, ActionGiveItem(oInventoryItem, oReceiver));
	}
}

// are bags passed as full bags, or traversed?
// inventory space limits?
// stacked items?
void GiveAllInventory(object oGiver, object oReceiver)
{
	//if ((GetObjectType(oGiver) != OBJECT_TYPE_CREATURE) && (GetObjectType(oReceiver) != OBJECT_TYPE_CREATURE))
	//{
	//	PrettyError("GiveAllInventory(): At least one of Giver or Receiver must be a creature.");		
	//	return;
	//}

    object oInventoryItem = GetFirstItemInInventory(oGiver);
    while (oInventoryItem != OBJECT_INVALID)
    {
		TransferItem(oGiver, oReceiver, oInventoryItem);
        //AssignCommand(oGiver, ActionGiveItem(oInventoryItem, oReceiver));
        oInventoryItem = GetNextItemInInventory(oGiver);
    }
}

// give gold, equipped items, and inventory
void GiveEverything(object oGiver, object oReceiver)
{
	if (GetObjectType(oGiver) != OBJECT_TYPE_CREATURE)
	{
		PrettyError("GiveEverything(): Everything Giver must be a creature!");		
		return;
	}

    int iGold = GetGold(oGiver);// Remove gold to crate
    AssignCommand(oReceiver, TakeGoldFromCreature(iGold, oGiver, FALSE));
	GiveAllEquippedItems(oGiver, oReceiver);
	GiveAllInventory(oGiver, oReceiver);
}

// Return number of items of Base Type in inventory
int CountInventoryItemsOfBaseType(int nBaseType, object oCreature=OBJECT_SELF)
{
    object oItem =  GetFirstItemInInventory(oCreature);
	int iItemCount=0;

    while (GetIsObjectValid(oItem))
    {
       if (GetBaseItemType(oItem) == nBaseType)
       {
			iItemCount++;
       }
       oItem =  GetNextItemInInventory(oCreature);
    }
    return iItemCount;
}


object GetInventoryItemOfBaseType(int nBaseType, int nNth=1, object oCreature=OBJECT_SELF)
{
    object oItem =  GetFirstItemInInventory(oCreature);
	int iItemCount=0;

    while (GetIsObjectValid(oItem))
    {
       if (GetBaseItemType(oItem) == nBaseType)
       {
			iItemCount++;
			if (iItemCount == nNth)
			{
				//PrettyDebug("Found requested item in inventory: " + GetName(oItem));
				return (oItem);
			}
       }
       oItem =  GetNextItemInInventory(oCreature);
    }
	
    return OBJECT_INVALID;
}


// list items in inventory.  Should work for placeables too.
int ListInventory(object oCreature=OBJECT_SELF)
{
    object oItem = GetFirstItemInInventory(oCreature);
	int i;

	PrettyDebug("*** Items in the inventory of " + GetName(oCreature));
    while (GetIsObjectValid(oItem))
    {
		i++;
		PrettyDebug("Item " + IntToString(i) + ": " + GetName(oItem));
       	oItem =  GetNextItemInInventory(oCreature);
    }
    return i;
}


// returns TRUE if an item was found
int EquipRandomArmorFromInventory()
{
	int bRet = FALSE;
	//PrettyDebug("Looking for something to change into");
	int iCount = CountInventoryItemsOfBaseType(BASE_ITEM_ARMOR);
	int iChoice = Random(iCount)+1;
	object oArmor = GetInventoryItemOfBaseType(BASE_ITEM_ARMOR, iChoice);
	if (GetIsObjectValid(oArmor))
	{
		bRet = TRUE;
		ActionUnequipItem(GetItemInSlot(INVENTORY_SLOT_CHEST));
		ActionEquipItem(oArmor, INVENTORY_SLOT_CHEST);
	}
	return (bRet);
}


// **** Functions for Identifying Item Types
	
// Is this a weapon of any type (including ammunition).
int GetIsWeapon(object oItem)
{
	return(IPGetIsMeleeWeapon(oItem) || IPGetIsRangedWeapon(oItem) || IPGetIsProjectile(oItem));
}

// is this a shield?
int GetIsShield(object oItem)
{
    int iType = GetBaseItemType(oItem);
    int iRet = FALSE;
    switch (iType)
    {
        case BASE_ITEM_SMALLSHIELD      :
        case BASE_ITEM_LARGESHIELD      :
        case BASE_ITEM_TOWERSHIELD      :
            iRet = TRUE;
            break;
	}
	return iRet;
}

// Is this armor or shield?
int GetIsArmorOrShield(object oItem)
{
    int iType = GetBaseItemType(oItem);
	return ((iType == BASE_ITEM_ARMOR) || GetIsShield(oItem));
}

// Is this a wonderous object
int GetIsWonderous(object oItem)
{
	int iRet = FALSE;
	int iType = GetBaseItemType(oItem);
	switch (iType)
	{
		case BASE_ITEM_AMULET			:
		case BASE_ITEM_BELT				:
		case BASE_ITEM_BOOTS			:
		case BASE_ITEM_BRACER			:
		case BASE_ITEM_CLOAK			:
		case BASE_ITEM_HELMET			:
		case BASE_ITEM_RING				:
			iRet = TRUE;
			break;
	}
	return iRet;
}

// Is item in specified (bit flag) category?
int GetIsItemCategory(object oItem, int iMatchCategory)
{
	int iItemCategory = GetItemCategory(oItem);
	//int iRet = GetState(iThisCategory, iItemCategory);
	//PrettyDebug(GetName(oItem) + " is in Category " + IntToString(iThisCategory) + 
	//	" In passed category of [" + IntToString(iItemCategory) + "] = " + IntToString(iRet));
	
	int bSwitch = FALSE;
	if (iMatchCategory == iItemCategory)
		bSwitch = TRUE;
	
	return bSwitch;
	
}

// Returns basic category which is includes weapon, armor, and other
int GetItemCategory(object oItem)
{
    int iRet;
	int iBaseType = GetBaseItemType(oItem);
	
	if (!GetIsObjectValid(oItem))
    	iRet = ITEM_CATEGORY_NONE;
	else if (GetIsWeapon(oItem)) 
		{
		if (iBaseType == BASE_ITEM_SHORTBOW || iBaseType == BASE_ITEM_LONGBOW)
			iRet = ITEM_CATEGORY_BOW;	
		else if (iBaseType == BASE_ITEM_HEAVYCROSSBOW || iBaseType == BASE_ITEM_LIGHTCROSSBOW)
			iRet = ITEM_CATEGORY_XBOW;
		else if (iBaseType == BASE_ITEM_SLING)
			iRet = ITEM_CATEGORY_SLING;
		else if (IPGetIsProjectile(oItem))
			iRet = ITEM_CATEGORY_AMMO;
		else
			iRet = ITEM_CATEGORY_WEAPON;
		}
	else if (GetIsArmorOrShield(oItem))
    	iRet = ITEM_CATEGORY_ARMOR_SHIELD;
	else if (GetIsWonderous(oItem))
		iRet = ITEM_CATEGORY_WONDER;
	else if (iBaseType == BASE_ITEM_GLOVES)
		iRet = ITEM_CATEGORY_GLOVES;
	else
		iRet = ITEM_CATEGORY_OTHER;
		
	//Gold Necklace override for Mephasm Charm
	if (GetTag(oItem) == "NW_IT_MNECK022")
		iRet = ITEM_CATEGORY_OTHER;

/*	
	//Now these are old masterwork system		
	//Handle new Masterwork weapons types
	if (iBaseType > 200 && iBaseType < 237)
		{
		if (iBaseType == 228)
			iRet = ITEM_CATEGORY_SLING;
		else if (iBaseType == 209 || iBaseType == 212)
			iRet = ITEM_CATEGORY_BOW;
		else if (iBaseType == 207 || iBaseType == 208)
			iRet = ITEM_CATEGORY_XBOW;
		else
			iRet = ITEM_CATEGORY_WEAPON;
		}
*/	
	
    return (iRet);
}


// return slot item is equipped in on creature.  -1 if not found
int GetSlotOfEquippedItem(object oItem, object oCreature)
{
	int iSlot;
	object oItemInSlot;
	
	for (iSlot=0; iSlot<NUM_INVENTORY_SLOTS; iSlot++)
	{
		oItemInSlot = GetItemInSlot(iSlot, oCreature);
		if (oItemInSlot == oItem)
		{
			return iSlot;
		}
	}
	return -1;
}

int GetIsCreatureSlot(int iSlot)
{
	int iRet = FALSE;
	if ((iSlot == INVENTORY_SLOT_CWEAPON_B) ||
		(iSlot == INVENTORY_SLOT_CWEAPON_L) ||
		(iSlot == INVENTORY_SLOT_CWEAPON_R) ||
		(iSlot == INVENTORY_SLOT_CARMOUR))
		iRet = TRUE;

	return (iRet);			
}

//===========================================================================
// Store related
//===========================================================================

// * returns a value that will be subtracted from the
// * oTarget's DC to resist APpraise or Persuasion
int N2_GetNPCEasyMark(object oTarget)
{
    int nCharmMod = 0;
    if (GetHasSpellEffect(SPELL_CHARM_PERSON, oTarget))
    {
        nCharmMod = 10;
    }
    else
    if (GetHasSpellEffect(SPELL_CHARM_MONSTER, oTarget))
    {
        nCharmMod = 10;
    }
    else
    if (GetHasSpellEffect(SPELL_CHARM_PERSON_OR_ANIMAL, oTarget))
    {
        nCharmMod = 10;
    }
    else if (GetHasSpellEffect(SPELL_MASS_CHARM, oTarget))
    {
            nCharmMod = 15;
    }
    else if (GetHasSpellEffect(SPELL_DOMINATE_MONSTER, oTarget))
    {
            nCharmMod = 20;
    }
    else if (GetHasSpellEffect(SPELL_DOMINATE_ANIMAL, oTarget))
    {
            nCharmMod = 20;
    }
    else if (GetHasSpellEffect(SPELL_DOMINATE_PERSON, oTarget))
    {
        nCharmMod = 20;
    }
    return nCharmMod;
}

// Open Store taking Appraise skill into account.
// changed from gplotAppraiseOpenStore() in NW_IO_PLOT - no longer random
/*
        An opposed skill ccomparison. Your appraise skill versus the shopkeepers appraise skill.
        Possible Results:
        Percentage Rebate/Penalty: The 'difference'
        Feedback: [Appraise Skill]: Merchant's reaction is unfavorable.
                  [Appraise Skill]: Merchant's reaction is neutral.
                  [Appraise Skill]: Merchant's reaction is favorable.
*/
void N2_AppraiseOpenStore(object oStore, object oPC, int nBonusMarkUp = 0, int nBonusMarkDown = 0)
{
    int STATE_FAILED = 1;
    int STATE_TIE = 2;
    int STATE_WON = 3;

    int nPlayerSkillRank = GetSkillRank(SKILL_APPRAISE, oPC);

    int nNPCSkillRank = GetSkillRank(SKILL_APPRAISE, OBJECT_SELF) - N2_GetNPCEasyMark(OBJECT_SELF);


    if (nNPCSkillRank < 1 )
        nNPCSkillRank = 1;


    int nState = 0;
	int nAdjust = nNPCSkillRank - nPlayerSkillRank; // * determines the level of price modification

	if (nNPCSkillRank > nPlayerSkillRank)
	{
	    nState = STATE_FAILED;
	}
	else
	if (nNPCSkillRank < nPlayerSkillRank)
	{
	    nState = STATE_WON;
	}
	else
	if (nNPCSkillRank == nPlayerSkillRank)
	{
	    nState = STATE_TIE;
	}

    if (nState == STATE_FAILED  )
    {
        FloatingTextStrRefOnCreature(182468, oPC, FALSE);
    }
    else
    if (nState == STATE_WON)
    {
        FloatingTextStrRefOnCreature(182470, oPC, FALSE);
    }
    else
    if (nState == STATE_TIE)
    {
        FloatingTextStrRefOnCreature(182469, oPC, FALSE);
    }


    // * Hard cap of 30% max up or down
    if (nAdjust > 30)
        nAdjust = 30;
    if (nAdjust < -30)
        nAdjust = -30;

	// nBonusMarkUp is added to the stores default mark up percentage on items sold (-100 to 100)
	// (positive is good for player)
	// Effect on selling items is only half of effect on buying items.
    nBonusMarkUp = nBonusMarkUp + nAdjust/2;
	
	// nBonusMarkDown is added to the stores default mark down percentage on items bought (-100 to 100)
	// (negative is good for player)
	nBonusMarkDown = nBonusMarkDown - nAdjust;
    OpenStore(oStore, oPC, nBonusMarkUp, nBonusMarkDown);
}